import argparse
import importlib
import logging
import sys

import ijson

from ocdskit.exceptions import CommandError

logger = logging.getLogger('ocdskit')

COMMAND_MODULES = (
    'ocdskit.cli.commands.combine_record_packages',
    'ocdskit.cli.commands.combine_release_packages',
    'ocdskit.cli.commands.compile',
    'ocdskit.cli.commands.convert_to_oc4ids',
    'ocdskit.cli.commands.detect_format',
    'ocdskit.cli.commands.echo',
    'ocdskit.cli.commands.indent',
    'ocdskit.cli.commands.mapping_sheet',
    'ocdskit.cli.commands.package_records',
    'ocdskit.cli.commands.package_releases',
    'ocdskit.cli.commands.schema_report',
    'ocdskit.cli.commands.schema_strict',
    'ocdskit.cli.commands.set_closed_codelist_enums',
    'ocdskit.cli.commands.split_record_packages',
    'ocdskit.cli.commands.split_release_packages',
    'ocdskit.cli.commands.tabulate',
    'ocdskit.cli.commands.upgrade',
    'ocdskit.cli.commands.validate',
)


def main():
    parser = argparse.ArgumentParser(description='Open Contracting Data Standard CLI')
    parser.add_argument('--encoding', help='the file encoding')
    parser.add_argument('--ascii', help='print escape sequences instead of UTF-8 characters', action='store_true')
    parser.add_argument('--pretty', help='pretty print output', action='store_true')

    subparsers = parser.add_subparsers(dest='subcommand')

    subcommands = {}

    for module in COMMAND_MODULES:
        try:
            command = importlib.import_module(module).Command(subparsers)
            subcommands[command.name] = command
        except ImportError as e:
            logger.error('exception "%s" prevented loading of %s module', e, module)

    args = parser.parse_args()

    if args.subcommand:
        command = subcommands[args.subcommand]
        try:
            command.args = args
            try:
                command.handle()
            except ijson.common.IncompleteJSONError as e:
                if str(e):
                    raise CommandError('JSON error: {}'.format(e))
                _raise_encoding_error(e, args.encoding)
            except UnicodeDecodeError as e:
                _raise_encoding_error(e, args.encoding)
        except CommandError as e:
            logger.critical(e)
            sys.exit(1)
    else:
        parser.print_help()


def _raise_encoding_error(e, encoding):
    if encoding and encoding.lower() == 'iso-8859-1':
        suggestion = 'utf-8'
    else:
        suggestion = 'iso-8859-1'
    raise CommandError('encoding error: {}\nTry `--encoding {}`?'.format(e, suggestion))


if __name__ == '__main__':
    main()



import json
from urllib.parse import urlparse

import requests
import rfc3987
from jsonschema import FormatChecker
from jsonschema.compat import str_types
from jsonschema.validators import Draft4Validator as validator

from ocdskit.cli.commands.base import OCDSCommand


class Command(OCDSCommand):
    name = 'validate'
    help = 'reads JSON data from standard input, validates it against the schema, and prints errors'

    def add_arguments(self):
        self.add_argument('--schema', help='the URL or path of the schema to validate against',
                          default='https://standard.open-contracting.org/latest/en/release-package-schema.json')
        self.add_argument('--check-urls', help='check the HTTP status code if "format": "uri"', action='store_true')
        self.add_argument('--timeout', help='timeout (seconds) to GET a URL', type=int, default=10)
        self.add_argument('--verbose', help='print items without validation errors', action='store_true')

    def handle(self):
        components = urlparse(self.args.schema)
        if components.scheme == 'file':
            with open(self.args.schema[7:]) as f:
                schema = json.load(f)
        else:
            schema = requests.get(self.args.schema).json()

        format_checker = FormatChecker()
        if self.args.check_urls:
            def check_url(instance):
                # See https://github.com/Julian/jsonschema/blob/master/jsonschema/_format.py
                if not isinstance(instance, str_types):
                    return True
                rfc3987.parse(instance, rule='URI')  # raises ValueError
                try:
                    response = requests.get(instance, timeout=self.args.timeout)
                    result = response.status_code in (200,)
                    if not result:
                        print('HTTP {} on GET {}'.format(response.status_code, instance))
                    return result
                except requests.exceptions.Timeout:
                    print('Timedout on GET {}'.format(instance))
                    return False

            format_checker.checks('uri', raises=(ValueError))(check_url)

        for i, data in enumerate(self.items()):
            errors = False
            for error in validator(schema, format_checker=format_checker).iter_errors(data):
                print('item {}: {} ({})'.format(i, error.message, '/'.join(error.absolute_schema_path)))
                errors = True
            if not errors and self.args.verbose:
                print('item {}: no errors'.format(i))



LANGUAGE_CODE_SUFFIX = '_(((([A-Za-z]{2,3}(-([A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-([A-Za-z]{4}))?(-([A-Za-z]{2}|[0-9]{3}))?(-([A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-([0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(x(-[A-Za-z0-9]{1,8})+))?)|(x(-[A-Za-z0-9]{1,8})+))'  # noqa: E501
LANGUAGE_CODE_SUFFIX_LEN = len(LANGUAGE_CODE_SUFFIX)


def _join_sep(sep, words):
    return sep + sep.join(words)


class Field:
    def __init__(self, schema=None, pointer=None, path=None, definition_pointer=None, definition_path=None,
                 required=None, deprecated=None, multilingual=None, sep='.'):
        """
        Initializes a schema field object.

        :param dict schema: the field's schema
        :param tuple pointer: the JSON pointer to the field in the schema, e.g.
                              ``('properties', 'tender', 'properties', 'id')``
        :param tuple path: the path to the field in data, e.g.
                           ``('tender', 'id')``
        :param tuple definition_pointer: the JSON pointer to the definition in which the field is defined, e.g.
                                         ``('definitions', 'Item')``
        :param tuple definition_path: the path to the definition in which the field is defined, e.g.
                                      ``('Item')``
        :param bool required: whether the field is listed in the schema's ``required``
        :param dict deprecated: if the field, or an ancestor of the field, sets ``deprecated``, the ``deprecated``
                                object
        :param bool multilingual: whether the field has a corresponding field in the schema's ``patternProperties``
        :param str sep: the separator to use in string representations of paths
        """
        self.schema = schema
        self.pointer_components = pointer
        self.path_components = path
        self.definition_pointer_components = definition_pointer
        self.definition_path_components = definition_path
        self.required = required
        self.deprecated = deprecated
        self.multilingual = multilingual
        self._sep = sep

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    @property
    def pointer(self):
        """
        Returns the JSON pointer to the field in the schema, e.g. ``/properties/tender/properties/id``.
        """
        return _join_sep('/', self.pointer_components)

    @property
    def definition_pointer(self):
        """
        Returns the JSON pointer to the definition in which the field is defined, e.g. ``/definitions/Item``.
        """
        return _join_sep('/', self.definition_pointer_components)

    @property
    def path(self):
        """
        Returns the path to the field in data with ``self.sep`` as separator, e.g. ``tender.id``.
        """
        return self._sep.join(self.path_components)

    @property
    def definition_path(self):
        """
        Returns the path to the definition in which the field is defined with ``self.sep`` as separator, e.g. ``Item``.
        """
        return self._sep.join(self.definition_path_components)

    @property
    def sep(self):
        """
        Returns the separator to use in string representations of paths.
        """
        return self._sep

    @sep.setter
    def sep(self, sep):
        """
        Sets the separator to use in string representations of paths.
        """
        self._sep = sep

    def __repr__(self):
        return repr(self.asdict())

    def asdict(self, sep=None, exclude=None):
        """
        Returns the field as a dict, with keys for: ``schema``, ``pointer``, ``path``,
        ``definition_pointer``, ``definition_path``, ``required``, ``deprecated``, ``multilingual``.

        :param list sep: the separator to use in string representations of paths, overriding ``self.sep``
        :param list exclude: a list of keys to exclude from the dict
        """
        d = {}

        exclude = exclude or ()
        sep = sep or self.sep

        for k, v in self.__dict__.items():
            if k not in exclude and not k.startswith('_') and not k.endswith('_components'):
                d[k] = v
        for k in ('pointer', 'definition_pointer'):
            if k not in exclude:
                d[k] = getattr(self, k)
        for k in ('path', 'definition_path'):
            if k not in exclude:
                d[k] = sep.join(getattr(self, '{}_components'.format(k)))

        return d


# This code is similar to `add_versioned` in `make_versioned_release_schema.py` in the `standard` repository.
def get_schema_fields(schema, pointer=None, path=None, definition_pointer=None, definition_path=None, deprecated=None):
    """
    Yields a ``Field`` object for each field (whether under ``properties`` or ``patternProperties``) in a JSON schema.

    :param dict schema: a JSON schema
    :param tuple pointer: the JSON pointer to the field in the schema, e.g.
                          ``('properties', 'tender', 'properties', 'id')``
    :param tuple path: the path to the field in data, e.g.
                       ``('tender', 'id')``
    :param tuple definition_pointer: the JSON pointer to the definition in which the field is defined, e.g.
                                     ``('definitions', 'Item')``
    :param tuple definition_path: the path to the definition in which the field is defined, e.g.
                                  ``('Item')``
    :param dict deprecated: if the field, or an ancestor of the field, sets ``deprecated``, the ``deprecated``
                            object
    """
    if pointer is None:
        pointer = ()
    if path is None:
        path = ()
    if definition_pointer is None:
        definition_pointer = ()
    if definition_path is None:
        definition_path = ()

    multilingual = set()
    hidden = set()
    for key, value in schema.get('patternProperties', {}).items():
        # The pattern might have an extra set of parentheses.
        for offset in (2, 1):
            end = -LANGUAGE_CODE_SUFFIX_LEN - offset
            # The pattern must be anchored and the suffix must occur at the end.
            if (key[end:-offset] == LANGUAGE_CODE_SUFFIX and
                    key[:offset] == '^('[:offset] and key[-offset:] == ')$'[-offset:]):
                multilingual.add(key[offset:end])
                hidden.add(key)
                break

    for key, value in schema.get('properties', {}).items():
        new_pointer = pointer + ('properties', key)
        new_path = path + (key,)
        required = schema.get('required', [])
        yield from _get_schema_field(key, value, new_pointer, new_path, definition_pointer, definition_path,
                                     required, deprecated or _deprecated(value), multilingual)

    for key, value in schema.get('definitions', {}).items():
        new_pointer = pointer + ('definitions', key)
        yield from get_schema_fields(value, pointer=new_pointer, path=(), definition_pointer=new_pointer,
                                     definition_path=(key,), deprecated=deprecated)

    for key, value in schema.get('patternProperties', {}).items():
        if key not in hidden:
            new_pointer = pointer + ('patternProperties', key)
            new_path = path + ('({})'.format(key),)
            yield Field(schema=value, pointer=new_pointer, path=new_path, definition_pointer=definition_pointer,
                        definition_path=definition_path, required=False, deprecated=deprecated or _deprecated(value),
                        multilingual=False)


def _get_schema_field(name, schema, pointer, path, definition_pointer, definition_path, required, deprecated,
                      multilingual):
    yield Field(schema=schema, pointer=pointer, path=path, definition_pointer=definition_pointer,
                definition_path=definition_path, required=name in required, deprecated=deprecated,
                multilingual=name in multilingual)

    if schema is None:
        return

    if 'properties' in schema or 'patternProperties' in schema:
        yield from get_schema_fields(schema, pointer=pointer, path=path, definition_pointer=definition_pointer,
                                     definition_path=definition_path, deprecated=deprecated)

    for key, value in schema.get('items', {}).get('properties', {}).items():
        new_pointer = pointer + ('items', 'properties', key)
        new_path = path + (key,)
        required = schema['items'].get('required', [])
        yield from _get_schema_field(key, value, new_pointer, new_path, definition_pointer, definition_path,
                                     required, deprecated or _deprecated(value), multilingual)


def _deprecated(value):
    if value is None:
        return {}
    return value.get('deprecated') or hasattr(value, '__reference__') and value.__reference__.get('deprecated') or {}



import logging
import os.path

import pytest

import ocdskit.packager

logger = logging.getLogger('vcr')
logger.setLevel(logging.WARNING)


@pytest.fixture()
def vcr_cassette_name(request):
    return '{}-{}'.format(os.path.splitext(os.path.basename(request.node.fspath))[0], request.node.name)


@pytest.fixture(params=[True, False])
def sqlite(request, monkeypatch):
    ocdskit.packager.using_sqlite = request.param



import itertools
import json
from collections import OrderedDict
from decimal import Decimal

from ocdsmerge.flatten import IdDict

try:
    import orjson
    jsonlib = orjson
    using_orjson = True
except ImportError:
    jsonlib = json
    using_orjson = False


# See `grouper` recipe: https://docs.python.org/3.8/library/itertools.html#recipes
def grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


# https://stackoverflow.com/questions/21663800/python-make-a-list-generator-json-serializable/46841935#46841935
class SerializableGenerator(list):
    def __init__(self, iterable):
        iterator = iter(iterable)
        try:
            # If `iter()` is omitted, then `__iter__` won't exhaust `head`.
            self.head = iter([next(iterator)])
            # Adding an item to the list ensures `__bool__` and `__len__` work.
            self.append(iterator)
        except StopIteration:
            # `__iter__` requires `head` to be set.
            self.head = []

    def __iter__(self):
        # `*self[:1]` is used, because `self[0]` raises IndexError when `iterable` is empty.
        return itertools.chain(self.head, *self[:1])


def _default(obj):
    if type(obj) in (IdDict, OrderedDict):  # needed by orjson, but not json
        return dict(obj)
    if isinstance(obj, Decimal):
        return float(obj)
    # https://docs.python.org/3/library/json.html#json.JSONEncoder.default
    try:
        iter(obj)
    except TypeError:
        pass
    else:
        return SerializableGenerator(obj)
    return json.JSONEncoder().default(obj)


def iterencode(data, ensure_ascii=False, **kwargs):
    """
    Returns a generator that yields each string representation as available.
    """
    if 'indent' not in kwargs:
        kwargs['separators'] = (',', ':')
    return json.JSONEncoder(ensure_ascii=ensure_ascii, default=_default, **kwargs).iterencode(data)


def json_dump(data, io, ensure_ascii=False, **kwargs):
    """
    Dumps JSON to a file-like object.
    """
    if 'indent' not in kwargs:
        kwargs['separators'] = (',', ':')
    json.dump(data, io, ensure_ascii=ensure_ascii, default=_default, **kwargs)


def json_dumps(data, ensure_ascii=False, **kwargs):
    """
    Dumps JSON to a string, and returns it.
    """
    # orjson doesn't support `ensure_ascii`, `indent` or `separators`.
    if not using_orjson or ensure_ascii or kwargs:
        if 'indent' not in kwargs:
            kwargs['separators'] = (',', ':')
        return json.dumps(data, default=_default, ensure_ascii=ensure_ascii, **kwargs)

    # orjson dumps to bytes.
    return orjson.dumps(data, default=_default).decode()


def get_ocds_minor_version(data):
    """
    Returns the OCDS minor version of the release package, record package or release.
    """
    if is_package(data):
        if 'version' in data:
            return data['version']
        return '1.0'
    else:  # release
        if 'parties' in data:
            return '1.1'
        return '1.0'


def is_package(data):
    """
    Returns whether the data is a record package or release package.
    """
    return is_record_package(data) or is_release_package(data)


def is_record_package(data):
    """
    Returns whether the data is a record package.
    """
    return 'records' in data


def is_record(data):
    """
    Returns whether the data is a record.
    """
    return 'releases' in data and 'ocid' in data


def is_release_package(data):
    """
    Returns whether the data is a release package.
    """
    return 'releases' in data and 'ocid' not in data


def is_release(data):
    """
    Returns whether the data is a release.
    """
    return 'tag' in data


def _empty_package(uri, publisher, published_date):
    if publisher is None:
        publisher = {}

    return {
        'uri': uri,
        'publisher': publisher,
        'publishedDate': published_date,
        'license': None,
        'publicationPolicy': None,
        'version': None,
        'extensions': {},
    }


def _empty_record_package(uri='', publisher=None, published_date=''):
    package = _empty_package(uri, publisher, published_date)
    package['packages'] = []
    package['records'] = []
    return package


def _empty_release_package(uri='', publisher=None, published_date=''):
    package = _empty_package(uri, publisher, published_date)
    package['releases'] = []
    return package


def _update_package_metadata(output, package):
    for field in ('publisher', 'license', 'publicationPolicy', 'version'):
        if field in package:
            output[field] = package[field]

    # We use an insertion-ordered dict to keep extensions in order without duplication.
    if 'extensions' in package:
        output['extensions'].update(dict.fromkeys(package['extensions']))


def _resolve_metadata(output, field):
    if output[field]:
        output[field] = list(output[field])
    else:
        del output[field]


def _remove_empty_optional_metadata(output):
    for field in ('license', 'publicationPolicy', 'version'):
        if output[field] is None:
            del output[field]



import os
import sys

import ijson

from ocdskit.util import iterencode, json_dumps


class StandardInputReader:
    def __init__(self, encoding):
        self.encoding = encoding

    def read(self, buf_size):
        data = sys.stdin.buffer.read(buf_size)
        if self.encoding is None or self.encoding == 'utf-8':
            return data
        return data.decode(self.encoding).encode('utf-8')


class BaseCommand:
    kwargs = {}

    def __init__(self, subparsers):
        """
        Initializes the subparser and adds arguments.
        """
        self.subparser = subparsers.add_parser(self.name, description=self.help, **self.kwargs)
        self.add_base_arguments()
        self.add_arguments()

    def add_base_arguments(self):
        """
        Adds default arguments to all commands.
        """
        pass

    def add_arguments(self):
        """
        Adds arguments specific to this command.
        """
        pass

    def add_argument(self, *args, **kwargs):
        """
        Adds an argument to the subparser.
        """
        self.subparser.add_argument(*args, **kwargs)

    def handle(self):
        """
        Runs the command.
        """
        raise NotImplementedError('commands must implement handle()')

    def prefix(self):
        """
        Returns the path to the items to process within each input.
        """
        return ''

    def items(self, **kwargs):
        """
        Yields the items in the input.
        """
        file = StandardInputReader(self.args.encoding)
        yield from ijson.items(file, self.prefix(), multiple_values=True, **kwargs)

    def print(self, data, streaming=False):
        """
        Prints JSON data.

        :param bool streaming: whether to stream output using ``json.JSONEncoder().iterencode()`` (it is only more
            memory efficient if ``data`` contains iterators)
        """
        kwargs = {}
        if self.args.pretty:
            kwargs['indent'] = 2
        if self.args.ascii:
            kwargs['ensure_ascii'] = True

        try:
            if streaming:
                for chunk in iterencode(data, **kwargs):
                    print(chunk, end='')
                print()
            else:
                print(json_dumps(data, **kwargs))
            sys.stdout.flush()
        # https://docs.python.org/3/library/signal.html#note-on-sigpipe
        except BrokenPipeError:
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())
            sys.exit(1)


class OCDSCommand(BaseCommand):
    def add_base_arguments(self):
        self.add_argument('--root-path', type=str, default='',
                          help='the path to the items to process within each input')

    def prefix(self):
        return self.args.root_path

    def items(self, **kwargs):
        """
        Yields the items in the input. If an item is an array, yields each entry of the array.
        """
        for item in super().items(**kwargs):
            if isinstance(item, list):
                for i in item:
                    yield i
            else:
                yield item

    def add_package_arguments(self, infix, prefix=''):
        """
        Adds arguments for setting package metadata to the subparser.
        """
        kwargs = {
            'infix': infix,
            'prefix': prefix,
        }

        template = "{prefix}set the {infix} package's {{}} to this value".format(**kwargs)

        self.add_argument('--uri', type=str, default='', help=template.format('uri'))
        self.add_argument('--published-date', type=str, default='', help=template.format('publishedDate'))
        self.add_argument('--publisher-name', type=str, default='', help=template.format("publisher's name"))
        self.add_argument('--publisher-uri', type=str, default='', help=template.format("publisher's uri"))
        self.add_argument('--publisher-scheme', type=str, default='', help=template.format("publisher's scheme"))
        self.add_argument('--publisher-uid', type=str, default='', help=template.format("publisher's uid"))
        self.add_argument('--fake', action='store_true',
                          help="{prefix}set the {infix} package's required metadata to dummy values".format(**kwargs))

    def parse_package_arguments(self):
        """
        Returns a publisher block as a dictionary.
        """
        metadata = {
            'uri': self.args.uri,
            'publisher': {},
            'published_date': self.args.published_date,
        }

        if self.args.fake:
            if not metadata['uri']:
                metadata['uri'] = 'placeholder:'
            if not metadata['published_date']:
                metadata['published_date'] = '9999-01-01T00:00:00Z'

        for key in ('publisher_name', 'publisher_uri', 'publisher_scheme', 'publisher_uid'):
            value = getattr(self.args, key)
            if value:
                metadata['publisher'][key[10:]] = value

        return metadata



import os.path
from tempfile import NamedTemporaryFile, TemporaryDirectory

from ocdskit.cli.__main__ import main
from tests import assert_command

content = b'{"lorem":"ipsum"}'
invalid = b'{"lorem":"ipsum"'


def test_command(monkeypatch):
    with NamedTemporaryFile() as f:
        f.write(content)
        f.flush()

        assert_command(monkeypatch, main, ['indent', f.name], '')
        f.seek(0)

        assert f.read() == b'{\n  "lorem": "ipsum"\n}\n'


def test_indent(monkeypatch):
    with NamedTemporaryFile() as f:
        f.write(content)
        f.flush()

        assert_command(monkeypatch, main, ['indent', '--indent', '4', f.name], '')
        f.seek(0)

        assert f.read() == b'{\n    "lorem": "ipsum"\n}\n'


def test_command_recursive(monkeypatch):
    with TemporaryDirectory() as d:
        with open(os.path.join(d, 'test.json'), 'wb') as f:
            f.write(content)
        with open(os.path.join(d, 'test.txt'), 'wb') as f:
            f.write(content)

        assert_command(monkeypatch, main, ['indent', '--recursive', d], '')

        with open(os.path.join(d, 'test.json'), 'rb') as f:
            assert f.read() == b'{\n  "lorem": "ipsum"\n}\n'
        with open(os.path.join(d, 'test.txt'), 'rb') as f:
            assert f.read() == content


def test_command_directory(monkeypatch, caplog):
    with TemporaryDirectory() as d:
        assert_command(monkeypatch, main, ['indent', d], '')

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'WARNING'
        assert caplog.records[0].message.endswith(' is a directory. Set --recursive to recurse into directories.')


def test_command_nonexistent(monkeypatch, caplog):
    assert_command(monkeypatch, main, ['indent', 'nonexistent'], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert caplog.records[0].message == 'nonexistent: No such file or directory'


def test_command_invalid_json(monkeypatch, caplog):
    with NamedTemporaryFile() as f:
        f.write(invalid)
        f.flush()

        assert_command(monkeypatch, main, ['indent', f.name], '')

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'ERROR'
        assert ' is not valid JSON. (json.decoder.JSONDecodeError: ' in caplog.records[0].message



import copy
import datetime
import decimal
import json
import logging
from collections import defaultdict

from jsonpointer import resolve_pointer
from ocdsmerge.util import sorted_releases

from ocdskit.combine import merge
from ocdskit.util import is_package

logger = logging.getLogger("ocdskit")


def check_type(item, item_type):
    """
    Check type and if incorrect return empty version of type so that future processing works with bad data.
    Should be used with dicts or lists that are then accessed later.
    """
    if not isinstance(item, item_type):
        if item:
            logger.warning("item {} is not of type {} so skipping".format(item, item_type.__name__))
        return item_type()
    return item


def cast_number_or_zero(item):
    """ Cast to decimal if fail return 0 so summing still works."""
    try:
        return decimal.Decimal(item)
    except (ValueError, TypeError):
        if item:
            logger.warning("item {} is not a number treating as zero".format(item))
        return 0


def cast_string(item):
    """
    Cast to string if possible. Does not try to convert dict, list, or None to string.
    Returns empty string on failure so future processing works.
    """

    if isinstance(item, (str, float, int, decimal.Decimal)):
        return str(item)

    if item:
        logger.warning("item {} is not able to be converted to a string".format(item))
    return ""


def run_transforms(config, releases, project_id=None, records=None, output=None):

    """
    Transforms a list of OCDS releases into a OC4IDS project.

    :param dict config: contains optional tranform options.
    :param list releases: list of OCDS releases or release packages
    :param string project_id: project ID of resulting project
    :param list records: pre computed list of records
    :param dict output: initial project output template project where transformed data will be added
    """
    transforms_to_run = []

    for transform in TRANSFORM_LIST:
        transform_name = transform.__name__
        if not config.get("all") and not config.get(transform_name) and transform_name in OPTIONAL_TRANSFORMS:
            continue
        transforms_to_run.append(transform)

    return _run_transforms(releases, project_id, records, output, transforms_to_run)


def _run_transforms(releases, project_id=None, records=None, output=None, transforms=None):

    state = InitialTransformState(releases, project_id, records, output)
    if not transforms:
        transforms = TRANSFORM_LIST

    for transform in transforms:
        transform(state)

    return state.output


class InitialTransformState:
    def __init__(self, releases_or_release_packages, project_id=None, records=None, output=None):

        # coerce generator into list as we iterate over it twice.
        releases_or_release_packages = list(releases_or_release_packages)
        all_releases = []
        for releases_or_release_package in releases_or_release_packages:
            if is_package(releases_or_release_package):
                all_releases.extend(check_type(releases_or_release_package.get("releases"), list))
            else:
                all_releases.append(releases_or_release_package)

        self.releases = sorted_releases(all_releases)
        self.releases_by_ocid = defaultdict(list)
        for release in self.releases:
            ocid = cast_string(release.get("ocid"))
            if ocid:
                self.releases_by_ocid[ocid].append(release)

        self.project_id = project_id
        records = records

        if not records:
            record_package = next(merge(releases_or_release_packages, return_package=True, use_linked_releases=True))
            records = check_type(record_package.get("records"), list)

        compiled_releases = []
        for record in records:
            compiled_release = check_type(record.get("compiledRelease"), dict)
            # projects only have linked releases 'uri' is a good proxy for that.
            record_releases = check_type(record.get("releases"), list)
            linked_releases = [release for release in record_releases if release.get("url")]
            embedded_releases = [release for release in record_releases if not release.get("url")]

            compiled_release["releases"] = linked_releases
            compiled_release["embeddedReleases"] = embedded_releases

            compiled_releases.append(compiled_release)

        self.compiled_releases = compiled_releases
        self.output = output or {}
        if project_id and "id" not in self.output:
            self.output["id"] = project_id

        self.party_analysis()

        self.generate_document_ids = False

    def party_analysis(self):

        all_parties = []
        party_ids = set()
        duplicate_party_ids = False

        for compiled_release in self.compiled_releases:
            parties = check_type(compiled_release.get("parties"), list)
            for party in parties:
                full_party_copy = copy.deepcopy(party)
                partial_party_copy = copy.deepcopy(party)
                partial_party_copy.pop("id", None)
                partial_party_copy.pop("roles", None)

                # A guess at the identity of the party. i.e id and roles missing
                party_fingerprint = json.dumps(partial_party_copy, sort_keys=True)

                unique_identifier = None
                identifier = check_type(party.get("identifier"), dict)
                id_ = cast_string(identifier.get("id"))
                scheme = cast_string(identifier.get("scheme"))
                if identifier and scheme:
                    unique_identifier = scheme + "-" + id_

                all_parties.append(
                    {
                        "party": full_party_copy,
                        "original_party": party,
                        "party_fingerprint": party_fingerprint,
                        "unique_identifier": unique_identifier,
                    }
                )
                party_id = party.get("id")
                if party_id in party_ids:
                    duplicate_party_ids = True
                party_ids.add(party_id)

        party_num = 1
        unique_parties = []

        for party in all_parties:
            found_party = None
            for unique_party in unique_parties:
                if party["unique_identifier"] and party["unique_identifier"] == unique_party["unique_identifier"]:
                    found_party = unique_party
                    break
                if party["party_fingerprint"] == unique_party["party_fingerprint"]:
                    found_party = unique_party
                    break
            if found_party:
                new_roles = set(
                    check_type(found_party["party"].get("roles"), list) + check_type(party["party"].get("roles"), list)
                )
                found_party["party"]["roles"] = list(new_roles)
            else:
                if duplicate_party_ids:
                    if party["unique_identifier"]:
                        party["party"]["id"] = party["unique_identifier"]
                    else:
                        party["party"]["id"] = str(party_num)
                        party_num += 1
                unique_parties.append(party)
            party["original_party"]["_new_id"] = party["party"]["id"]

        self.parties = [party["party"] for party in unique_parties]


def copy_party_to_party_list(state, party):
    output_parties = state.output.get("parties", [])
    if not output_parties:
        state.output["parties"] = output_parties

    output_party = None
    for party_to_match in output_parties:
        if party.get("id") == party_to_match.get("id"):
            output_party = party_to_match

    if not output_party:
        output_party = copy.deepcopy(party)
        output_parties.append(output_party)
    return output_party


def copy_party_by_role(state, role, new_roles=None):
    for party in state.parties:
        if role in check_type(party.get("roles"), list):
            output_party = copy_party_to_party_list(state, party)
            if new_roles:
                output_roles = output_party.get("roles", [])
                output_roles.extend(new_roles)
                output_party["roles"] = list(set(output_roles))


def copy_document(state, document):
    """
    Copies a document. If it finds clashing ids change ids to autoincrement numbers
    """

    output_documents = state.output.get("documents")
    if not output_documents:
        output_documents = []
        state.output["documents"] = output_documents

    duplicate_doc_id = False

    for output_document in output_documents:
        if output_document.get("id") == document.get("id"):
            duplicate_doc_id = True
            break

    output_documents.append(document)

    if duplicate_doc_id:
        for num, doc in enumerate(output_documents):
            doc["id"] = str(num + 1)


def copy_document_by_type(state, document_type):

    """
    Copies documents of specific documentType from planning.documents to documents
    """

    for compiled_release in state.compiled_releases:
        documents = resolve_pointer(compiled_release, "/planning/documents", [])
        for document in check_type(documents, list):
            document = check_type(document, dict)
            if document_type == document.get("documentType"):
                copy_document(state, document)


def concat_ocid_and_string(state, path_to_string):

    """
    Places the ocid of a release in front of a string (eg. description or title)
    so that it can be joined unambiguously with others, separated by new lines
    """

    strings = ""
    for compiled_release in state.compiled_releases:

        ocid = cast_string(resolve_pointer(compiled_release, "/ocid", None))
        a_string = cast_string(resolve_pointer(compiled_release, path_to_string, None))

        if a_string:
            concat = "<{}> {}\n".format(ocid, a_string)
            strings = strings + concat

    return strings


def public_authority_role(state):
    """
    CoST IDS element: Project owner
    """
    copy_party_by_role(state, "publicAuthority")


def buyer_role(state):
    """
    CoST IDS element: Project owner
    """
    copy_party_by_role(state, "buyer", ["publicAuthority"])


def sector(state):
    """
    CoST IDS element: Sector
    """
    sectors = []
    for compiled_release in state.compiled_releases:
        sector_id = cast_string(resolve_pointer(compiled_release, "/planning/project/sector/id", ""))
        sector_scheme = cast_string(resolve_pointer(compiled_release, "/planning/project/sector/scheme", ""))
        if sector_scheme:
            sector_name = sector_scheme + "-" + sector_id
        else:
            sector_name = sector_id

        sectors.append(sector_name)

    if sectors:
        state.output["sector"] = list(set(sectors))


def additional_classifications(state):
    """
    CoST IDS element: Subsector
    """
    for compiled_release in state.compiled_releases:
        additional_classifications = resolve_pointer(
            compiled_release, "/planning/project/additionalClassifications", None
        )
        if additional_classifications:
            output_classifications = state.output.get("additionalClassifications", [])
            state.output["additionalClassifications"] = output_classifications
            for classification in additional_classifications:
                if classification not in output_classifications:
                    output_classifications.append(classification)


def title(state):
    """
    CoST IDS element: Project name
    """
    found_title = None
    for compiled_release in state.compiled_releases:
        title = resolve_pointer(compiled_release, "/planning/project/title", None)
        if title:
            if found_title and found_title != title:
                logger.warning("Multiple differing titles found for project {}".format(state.project_id))
                return
        found_title = title
    if found_title:
        state.output["title"] = title


def title_from_tender(state):
    """
    CoST IDS element: Project name
    """
    if state.output.get("title"):
        return

    titles = []
    for compiled_release in state.compiled_releases:
        title = resolve_pointer(compiled_release, "/tender/title", None)
        if title:
            titles.append(title)
    if len(titles) > 1:
        state.output["title"] = concat_ocid_and_string(state, "/tender/title")
    elif len(titles) == 1:
        state.output["title"] = titles[0]


def contracting_process_setup(state):
    """ This will initailly create the contracting process objects and the summary object
    within.  All transforms that use contracting processes need to run this tranform first."""

    state.output["contractingProcesses"] = []

    for compiled_release in state.compiled_releases:
        contracting_process = {}
        contracting_process["id"] = compiled_release.get("ocid")
        contracting_process["summary"] = {}
        contracting_process["summary"]["ocid"] = compiled_release.get("ocid")

        releases = compiled_release.get("releases")
        if releases:
            contracting_process["releases"] = releases

        embedded_releases = compiled_release.get("embeddedReleases")
        if embedded_releases:
            contracting_process["embeddedReleases"] = embedded_releases

        state.output["contractingProcesses"].append(contracting_process)


def procuring_entity(state):
    """
    CoST IDS element: Procuring entity
    """
    copy_party_by_role(state, "procuringEntity")

    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        procuring_entities = []
        for party in check_type(compiled_release.get("parties"), list):
            if "procuringEntity" in check_type(party.get("roles"), list):
                procuring_entities.append(party)
        if len(procuring_entities) > 1:
            logger.warning(
                "More than one procuringEntity in contractingProcesses with ocid {} skipping tranform".format(
                    compiled_release.get("ocid")
                )
            )
            continue
        if procuring_entities:
            tender = check_type(contracting_process["summary"].get("tender"), dict)
            procuring_entity = {}
            procuring_entity["id"] = procuring_entities[0].get("_new_id")
            name = procuring_entities[0].get("name")
            if name:
                procuring_entity["name"] = name
            tender["procuringEntity"] = procuring_entity
            contracting_process["summary"]["tender"] = tender


def administrative_entity(state):
    """
    CoST IDS element: Contract administrative entity
    """
    copy_party_by_role(state, "administrativeEntity")

    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        administrative_entities = []
        for party in check_type(compiled_release.get("parties"), list):
            if "administrativeEntity" in check_type(party.get("roles"), list):
                administrative_entities.append(party)
        if len(administrative_entities) > 1:
            logger.warning(
                "More than one administrativeEntity in contractingProcesses with ocid {} skipping tranform".format(
                    compiled_release.get("ocid")
                )
            )
            continue
        if administrative_entities:
            tender = check_type(contracting_process["summary"].get("tender"), dict)
            administrative_entity = {}
            administrative_entity["id"] = administrative_entities[0].get("_new_id")
            administrative_entity["name"] = administrative_entities[0].get("name")
            tender["administrativeEntity"] = administrative_entity
            contracting_process["summary"]["tender"] = tender


def contract_status(state):
    """
    CoST IDS element: Contract status
    """
    current_iso_datetime = datetime.datetime.now().isoformat()

    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        tender = check_type(compiled_release.get("tender"), dict)
        tender_status = tender.get("status")
        closed_tender = tender_status in ("cancelled", "unsuccessful", "withdrawn")
        contracts = check_type(compiled_release.get("contracts"), list)
        awards = check_type(compiled_release.get("awards"), list)

        contract_periods = []
        if tender:
            tender_contract_period = check_type(tender.get("contractPeriod"), dict)
            if tender_contract_period:
                contract_periods.append(tender_contract_period)
        for contract in contracts:
            contract_contract_period = check_type(contract.get("period"), dict)
            if contract_contract_period:
                contract_periods.append(contract_contract_period)
        for award in awards:
            award_contract_period = check_type(award.get("contractPeriod"), dict)
            if award_contract_period:
                contract_periods.append(award_contract_period)

        # pre-award
        if tender and not closed_tender:
            if not compiled_release.get("awards") and not compiled_release.get("contracts"):
                contracting_process["summary"]["status"] = "pre-award"
                continue

            all_contracts_pending = all((contract.get("status") == "pending") for contract in contracts)
            all_awards_pending = all(award.get("status") == "pending" for award in awards)

            if all_contracts_pending and all_awards_pending:
                contracting_process["summary"]["status"] = "pre-award"
                continue

            all_awards_in_future = all(cast_string(award.get("date")) > current_iso_datetime for award in awards)

            award_period_start_date = cast_string(resolve_pointer(tender, "/awardPeriod/startDate", ""))
            award_period_in_future = award_period_start_date > current_iso_datetime

            if all_awards_in_future and award_period_in_future:
                contracting_process["summary"]["status"] = "pre-award"
                continue

        # active

        if any(contract.get("status") == "active" for contract in contracts):
            contracting_process["summary"]["status"] = "active"
            continue

        if any(
            (cast_string(period.get("startDate")) < current_iso_datetime)
            and (current_iso_datetime < cast_string(period.get("endDate", "9999-12-31")))
            for period in contract_periods
            if period
        ):
            contracting_process["summary"]["status"] = "active"
            continue

        # closed

        if closed_tender:
            contracting_process["summary"]["status"] = "closed"
            continue

        if awards:
            if all(check_type(award, dict).get("status") in ("cancelled", "unsuccessful") for award in awards):
                contracting_process["summary"]["status"] = "closed"
                continue

        if contracts:
            if all(check_type(contract, dict).get("status") in ("cancelled", "terminated") for contract in contracts):
                contracting_process["summary"]["status"] = "closed"
                continue

        if all(current_iso_datetime > cast_string(period.get("endDate", "9999-12-31")) for period in contract_periods):
            contracting_process["summary"]["status"] = "closed"
            continue


def procurement_process(state):
    """
    CoST IDS element: Procurement process
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        input_tender = check_type(compiled_release.get("tender"), dict)
        if input_tender:
            procurement_method = input_tender.get("procurementMethod")
            if procurement_method:
                tender = contracting_process["summary"].get("tender", {})
                tender["procurementMethod"] = procurement_method
                contracting_process["summary"]["tender"] = tender

            procurement_method_details = input_tender.get("procurementMethodDetails")
            if procurement_method_details:
                tender = contracting_process["summary"].get("tender", {})
                tender["procurementMethodDetails"] = procurement_method_details
                contracting_process["summary"]["tender"] = tender


def number_of_tenderers(state):
    """
    CoST IDS element: Number of firms tendering
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        input_tender = check_type(compiled_release.get("tender"), dict)
        if input_tender:
            number_of_tenderers = input_tender.get("numberOfTenderers")
            if number_of_tenderers:
                tender = contracting_process["summary"].get("tender", {})
                tender["numberOfTenderers"] = number_of_tenderers
                contracting_process["summary"]["tender"] = tender


def location(state):
    """
    CoST IDS element: Project location
    """
    all_locations = []
    for compiled_release in state.compiled_releases:
        locations = resolve_pointer(compiled_release, "/planning/project/locations", None)
        if locations:
            all_locations.extend(locations)

    if all_locations:
        state.output["locations"] = all_locations


def location_from_items(state):
    """
    CoST IDS element: Project location
    """
    if state.output.get("locations"):
        return

    locations = []
    for compiled_release in state.compiled_releases:

        items = resolve_pointer(compiled_release, "/tender/items", [])
        for item in check_type(items, list):

            delivery_location = resolve_pointer(item, "/deliveryLocation", None)
            if delivery_location:
                locations.append(delivery_location)

            delivery_address = resolve_pointer(item, "/deliveryAddress", None)
            if delivery_address:
                locations.append({"address": delivery_address})

        if len(locations) > 0:
            state.output["locations"] = locations
            break


def budget(state):
    """
    CoST IDS element: Budget
    """

    if len(state.compiled_releases) == 1:
        budget_value = resolve_pointer(state.compiled_releases[0], "/planning/budget/amount", None)
        if budget_value:
            state.output["budget"] = {"amount": budget_value}
    else:
        budget_currencies = set()
        budget_amounts = []

        for compiled_release in state.compiled_releases:
            budget_amounts.append(
                cast_number_or_zero(resolve_pointer(compiled_release, "/planning/budget/amount/amount", None))
            )
            budget_currencies.add(resolve_pointer(compiled_release, "/planning/budget/amount/currency", None))

        if len(budget_currencies) > 1:
            logger.warning("Can't get budget total, {} different currencies found.".format(len(budget_currencies)))
        else:
            state.output["budget"] = {
                "amount": {"amount": sum(budget_amounts), "currency": next(iter(budget_currencies))}
            }


def budget_approval(state):
    """
    CoST IDS element: Project budget approval date
    """
    copy_document_by_type(state, "budgetApproval")


def environmental_impact(state):
    """
    CoST IDS element: Environmental impact
    """
    copy_document_by_type(state, "environmentalImpact")


def land_and_settlement_impact(state):
    """
    CoST IDS element: Land and settlement impact
    """
    copy_document_by_type(state, "landAndSettlementImpact")


def purpose(state):
    """
    CoST IDS element: Purpose
    """
    if len(state.compiled_releases) == 1:
        rationale = resolve_pointer(state.compiled_releases[0], "/planning/rationale", None)
        if rationale:
            state.output["purpose"] = rationale
    else:
        purposes = concat_ocid_and_string(state, "/planning/rationale")
        if purposes != "":
            state.output["purpose"] = purposes


def purpose_needs_assessment(state):
    """
    CoST IDS element: Purpose
    """
    copy_document_by_type(state, "needsAssessment")


def description(state):
    """
    CoST IDS element: Project description
    """
    output_description = None

    for compiled_release in state.compiled_releases:
        description = resolve_pointer(compiled_release, "/planning/project/description", None)
        if description:
            if output_description and output_description != description:
                logger.warning(
                    "Multiple differing planning/project/descriptions found eg. {}, {}, skipping conversion".format(
                        description, output_description
                    )
                )
                return
            output_description = description

    if output_description:
        state.output["description"] = output_description


def description_tender(state):
    """
    CoST IDS element: Project description
    """
    if state.output.get("description"):
        return

    if len(state.compiled_releases) == 1:
        description = resolve_pointer(state.compiled_releases[0], "/tender/description", None)
        if description:
            state.output["description"] = description
            return

    else:
        descriptions = concat_ocid_and_string(state, "/tender/description")
        if descriptions != "":
            state.output["description"] = descriptions
            return


def funding_sources(state):
    """
    CoST IDS element: Funding sources
    """

    found_funding_source = False
    for compiled_release in state.compiled_releases:

        parties = check_type(resolve_pointer(compiled_release, "/parties", None), list)

        # Get parties from budgetBreakdown.sourceParty
        breakdowns = check_type(resolve_pointer(compiled_release, "/planning/budget/budgetBreakdown", None), list)
        if breakdowns:
            for breakdown in breakdowns:
                source_party = check_type(resolve_pointer(breakdown, "/sourceParty", None), dict)
                party_id = source_party.get("id")
                # Look up party data by id in parties
                if parties and party_id:
                    for party in parties:
                        party = check_type(party, dict)
                        if party.get("id") == party_id:
                            output_party = copy.deepcopy(party)
                            # Add to parties and set funder in roles
                            if check_type(output_party.get("roles"), list):
                                output_party["roles"].append("funder")
                            else:
                                output_party["roles"] = ["funder"]
                            output_party["id"] = output_party.pop("_new_id")
                            copy_party_to_party_list(state, output_party)
                            found_funding_source = True

    # If no parties from the budget breakdown, copy from top level with 'funder' roles
    if not found_funding_source:
        copy_party_by_role(state, "funder")


def cost_estimate(state):
    """
    CoST IDS element: Cost estimate
    """
    for contracting_process in state.output["contractingProcesses"]:
        ocid = contracting_process.get("id")
        latest_planning_value = None
        for release in state.releases_by_ocid.get(ocid, []):
            tender_status = resolve_pointer(release, "/tender/status", None)
            tender_value = resolve_pointer(release, "/tender/value", None)
            if tender_status == "planning" and tender_value:
                latest_planning_value = tender_value

        if latest_planning_value:
            tender = contracting_process["summary"].get("tender", {})
            tender["costEstimate"] = latest_planning_value
            contracting_process["summary"]["tender"] = tender


def contract_title(state):
    """
    CoST IDS element: Contract title
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        contracts = check_type(compiled_release.get("contracts"), list)
        awards = check_type(compiled_release.get("awards"), list)

        if len(contracts) > 1 or len(awards) > 1:
            tender_title = resolve_pointer(compiled_release, "/tender/title", None)

            if tender_title:
                contracting_process["summary"]["title"] = tender_title
            continue

        if len(contracts) == 1:
            contract = check_type(contracts[0], dict)
            contract_title = contract.get("title")
            if contract_title:
                contracting_process["summary"]["title"] = contract_title
            continue

        if len(awards) == 1:
            award = check_type(awards[0], dict)
            award_title = award.get("title")
            if award_title:
                contracting_process["summary"]["title"] = award_title
            continue


def suppliers(state):
    """
    CoST IDS element: Contract firm(s)
    """
    copy_party_by_role(state, "supplier")

    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        parties = check_type(compiled_release.get("parties"), list)
        suppliers = []
        for party in parties:
            if "supplier" in check_type(party.get("roles"), list):
                suppliers.append({"id": party.get("id"), "name": party.get("name")})

        if suppliers:
            contracting_process["summary"]["suppliers"] = suppliers


def contract_price(state):
    """
    CoST IDS element: Contract price
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        awards = check_type(compiled_release.get("awards"), list)
        award_currency = None
        award_amount = 0
        for award in awards:
            award = check_type(award, dict)
            amount = cast_number_or_zero(resolve_pointer(award, "/value/amount", None))
            award_amount += amount

            currency = cast_string(resolve_pointer(award, "/value/currency", None))
            if currency:
                if award_currency is None:
                    award_currency = currency
                else:
                    if currency != award_currency:
                        logger.warning("Multiple currencies not supported {}, {}".format(award_currency, currency))
                        award_amount = 0
                        break

        if award_amount:
            contracting_process["summary"]["contractValue"] = {"amount": award_amount, "currency": award_currency}


def contract_process_description(state):
    """
    CoST IDS element: Contract scope of work
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        contracts = check_type(compiled_release.get("contracts"), list)
        awards = check_type(compiled_release.get("awards"), list)

        if len(contracts) > 1 or len(awards) > 1:
            tender = check_type(compiled_release.get("tender"), dict)
            tender_description = tender.get("description")
            if tender_description:
                contracting_process["summary"]["description"] = tender_description
                continue

            tender_items_descriptions = []
            for tender_item in check_type(tender.get("items"), list):
                tender_item = check_type(tender_item, dict)
                tender_item_description = tender_item.get("description")
                if tender_item_description:
                    tender_items_descriptions.append(tender_item_description)

            if len(tender_items_descriptions) == 1:
                contracting_process["summary"]["description"] = tender_items_descriptions[0]
            continue

        contract_descriptions = []
        contract_items_descriptions = []
        for contract in contracts:
            contract = check_type(contract, dict)
            contract_description = contract.get("description")
            if contract_description:
                contract_descriptions.append(contract_description)
            for contract_item in check_type(contract.get("items"), list):
                contract_item = check_type(contract_item, dict)
                contract_item_description = contract_item.get("description")
                if contract_item_description:
                    contract_items_descriptions.append(contract_item_description)

        if len(contract_descriptions) == 1:
            contracting_process["summary"]["description"] = contract_descriptions[0]
            continue
        if len(contract_items_descriptions) == 1:
            contracting_process["summary"]["description"] = contract_items_descriptions[0]
            continue

        award_descriptions = []
        award_items_descriptions = []
        for award in awards:
            award = check_type(award, dict)
            award_description = award.get("description")
            if award_description:
                award_descriptions.append(award_description)
            for award_item in check_type(award.get("items"), list):
                award_item = check_type(award_item, dict)
                award_item_description = award_item.get("description")
                if award_item_description:
                    award_items_descriptions.append(award_item_description)

        if len(award_descriptions) == 1:
            contracting_process["summary"]["description"] = award_descriptions[0]
            continue
        if len(award_items_descriptions) == 1:
            contracting_process["summary"]["description"] = award_items_descriptions[0]
            continue


def contract_period(state):
    """
    CoST IDS element: Contract start date and contract period (duration)
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):
        awards = check_type(compiled_release.get("awards"), list)
        start_dates = []
        end_dates = []
        for award in awards:
            contract_period = check_type(award.get("contractPeriod"), dict)
            start_date = cast_string(contract_period.get("startDate"))
            if start_date:
                start_dates.append(start_date)
            end_date = cast_string(contract_period.get("endDate"))
            if end_date:
                end_dates.append(end_date)
        if start_dates and end_dates:
            contracting_process["summary"]["contractPeriod"] = {
                "startDate": min(start_dates),
                "endDate": max(end_dates),
            }
            continue

        tender = check_type(compiled_release.get("tender"), dict)
        contract_period = tender.get("contractPeriod")
        if contract_period:
            contracting_process["summary"]["contractPeriod"] = contract_period


def project_scope(state):
    """
    CoST IDS element: Project Scope (main output) and Project Scope (projected)
    """
    copy_document_by_type(state, "projectScope")


def project_scope_summary(state):
    """
    CoST IDS element: Project Scope (main output)
    """
    for compiled_release, contracting_process in zip(state.compiled_releases, state.output["contractingProcesses"]):

        release_tender = compiled_release.get("tender", {})
        items = check_type(release_tender.get("items"), list)
        milestones = check_type(release_tender.get("milestones"), list)

        tender = contracting_process["summary"].get("tender", {})

        if items:
            tender.setdefault("items", []).extend(items)
        if milestones:
            tender.setdefault("milestones", []).extend(milestones)

        contracting_process["summary"]["tender"] = tender


def final_audit(state):
    """
    CoST IDS element: Reference to audit and evaluation reports
    """
    if not state.output.get("documents"):
        state.output["documents"] = []

    for compiled_release in state.compiled_releases:

        contracts = resolve_pointer(compiled_release, "/contracts", [])
        for contract in contracts:
            documents = resolve_pointer(contract, "/implementation/documents", [])
            for document in check_type(documents, list):
                document = check_type(document, dict)
                if document.get("documentType") == "finalAudit":
                    state.output["documents"].append(document)


TRANSFORM_LIST = [
    contracting_process_setup,
    public_authority_role,
    buyer_role,
    sector,
    additional_classifications,
    title,
    title_from_tender,
    procuring_entity,
    administrative_entity,
    contract_status,
    procurement_process,
    location,
    location_from_items,
    budget,
    budget_approval,
    purpose,
    purpose_needs_assessment,
    description,
    description_tender,
    environmental_impact,
    land_and_settlement_impact,
    funding_sources,
    cost_estimate,
    contract_title,
    suppliers,
    contract_price,
    contract_process_description,
    contract_period,
    project_scope,
    project_scope_summary,
    final_audit,
]


OPTIONAL_TRANSFORMS = [
    "buyer_role",
    "title_from_tender",
    "location_from_items",
    "purpose_needs_assessment",
    "description_tender",
    "project_scope_summary",
]



import warnings

from ocdsextensionregistry import ProfileBuilder
from ocdsmerge import Merger
from ocdsmerge.util import get_release_schema_url, get_tags

from ocdskit.packager import Packager
from ocdskit.util import (_empty_record_package, _empty_release_package, _remove_empty_optional_metadata,
                          _resolve_metadata, _update_package_metadata)


def _package(key, items, uri, publisher, published_date, extensions):
    if publisher is None:
        publisher = {}
    if 'name' not in publisher:
        publisher['name'] = ''
    if extensions is None:
        extensions = []

    output = {
        'uri': uri,
        'publisher': publisher,
        'publishedDate': published_date,
        'version': '1.1',  # fields might be deprecated
        'extensions': extensions,
        key: items,
    }

    return output


def package_records(records, uri='', publisher=None, published_date='', extensions=None):
    """
    Wraps records in a record package.

    :param list records: a list of records
    :param str uri: the record package's ``uri``
    :param dict publisher: the record package's ``publisher``
    :param str published_date: the record package's ``publishedDate``
    :param list extensions: the record package's ``extensions``
    """
    return _package('records', records, uri, publisher, published_date, extensions)


def package_releases(releases, uri='', publisher=None, published_date='', extensions=None):
    """
    Wraps releases in a release package.

    :param list releases: a list of releases
    :param str uri: the release package's ``uri``
    :param dict publisher: the release package's ``publisher``
    :param str published_date: the release package's ``publishedDate``
    :param list extensions: the release package's ``extensions``
    """
    return _package('releases', releases, uri, publisher, published_date, extensions)


def combine_record_packages(packages, uri='', publisher=None, published_date=''):
    """
    Collects the packages and records from the record packages into one record package.

    :param packages: an iterable of record packages
    :param str uri: the record package's ``uri``
    :param dict publisher: the record package's ``publisher``
    :param str published_date: the record package's ``publishedDate``
    """
    # See options for not buffering all inputs into memory: https://github.com/open-contracting/ocdskit/issues/119
    output = _empty_record_package(uri, publisher, published_date)
    output['packages'] = {}

    for package in packages:
        _update_package_metadata(output, package)
        if 'records' in package:
            output['records'].extend(package['records'])
        if 'packages' in package:
            output['packages'].update(dict.fromkeys(package['packages']))

    if publisher:
        output['publisher'] = publisher

    _resolve_metadata(output, 'packages')
    _resolve_metadata(output, 'extensions')
    _remove_empty_optional_metadata(output)

    return output


def combine_release_packages(packages, uri='', publisher=None, published_date=''):
    """
    Collects the releases from the release packages into one release package.

    :param packages: an iterable of release packages
    :param str uri: the release package's ``uri``
    :param dict publisher: the release package's ``publisher``
    :param str published_date: the release package's ``publishedDate``
    """
    # See options for not buffering all inputs into memory: https://github.com/open-contracting/ocdskit/issues/119
    output = _empty_release_package(uri, publisher, published_date)

    for package in packages:
        _update_package_metadata(output, package)
        if 'releases' in package:
            output['releases'].extend(package['releases'])

    if publisher:
        output['publisher'] = publisher

    _resolve_metadata(output, 'extensions')
    _remove_empty_optional_metadata(output)

    return output


def merge(data, uri='', publisher=None, published_date='', schema=None, return_versioned_release=False,
          return_package=False, use_linked_releases=False, streaming=False):
    """
    Merges release packages and individual releases.

    By default, yields compiled releases. If ``return_versioned_release`` is ``True``, yields versioned releases. If
    ``return_package`` is ``True``, wraps the compiled releases (and versioned releases if ``return_versioned_release``
    is ``True``) in a record package.

    If ``return_package`` is set and ``publisher`` isn't set, the output record package will have the same publisher as
    the last input release package.

    :param data: an iterable of release packages and individual releases
    :param str uri: if ``return_package`` is ``True``, the record package's ``uri``
    :param dict publisher: if ``return_package`` is ``True``, the record package's ``publisher``
    :param str published_date: if ``return_package`` is ``True``, the record package's ``publishedDate``
    :param dict schema: the URL, path or dict of the patched release schema to use
    :param bool return_package: wrap the compiled releases in a record package
    :param bool use_linked_releases: if ``return_package`` is ``True``, use linked releases instead of full releases,
        if the input is a release package
    :param bool return_versioned_release: if ``return_package`` is ``True``, include versioned releases in the record
        package; otherwise, yield versioned releases instead of compiled releases
    :param bool streaming: if ``return_package`` is ``True``, set the package's records to a generator (this only works
        if the calling code exhausts the generator before ``merge`` returns)
    """
    with Packager() as packager:
        packager.add(data)

        if not schema and packager.version:
            prefix = packager.version.replace('.', '__') + '__'
            tag = next(tag for tag in reversed(get_tags()) if tag.startswith(prefix))
            schema = get_release_schema_url(tag)

            if packager.package['extensions']:
                builder = ProfileBuilder(tag, list(packager.package['extensions']))
                schema = builder.patched_release_schema()

        merger = Merger(schema)

        if return_package:
            packager.package['uri'] = uri
            packager.package['publishedDate'] = published_date
            if publisher:
                packager.package['publisher'] = publisher

            yield from packager.output_package(merger, return_versioned_release=return_versioned_release,
                                               use_linked_releases=use_linked_releases, streaming=streaming)
        else:
            yield from packager.output_releases(merger, return_versioned_release=return_versioned_release)


def compile_release_packages(*args, **kwargs):
    warnings.warn('compile_release_packages() is deprecated. Use merge() instead.', DeprecationWarning, stacklevel=2)
    yield from merge(*args, **kwargs)



from ocdskit.cli.__main__ import main
from tests import assert_streaming


def test_command(monkeypatch):
    assert_streaming(monkeypatch, main, ['convert-to-oc4ids', '--project-id', '1'],
                     ['release_1.1.json'],
                     ['oc4ids-project_minimal.json'])



import logging

from ocdskit import oc4ids
from ocdskit.cli.commands.base import OCDSCommand

logger = logging.getLogger("ocdskit")


class Command(OCDSCommand):
    name = "convert-to-oc4ids"
    help = '''reads individual releases or release packages from standard input,
              and prints a single project conforming to the OC4IDS'''

    def add_arguments(self):
        self.add_argument("--project-id", help="set the project's id to this value")
        self.add_argument("--all-transforms", help="run all optional transforms", action="store_true")
        self.add_argument("--transforms", help="comma-separated list of optional transforms to run", default="")
        self.add_argument("--package", action="store_true", help="wrap the project in a package")

        self.add_package_arguments("project", "if --package is set, ")

    def handle(self):
        project_id = self.args.project_id
        config = {}

        if self.args.all_transforms:
            config["all"] = True
        else:
            for option in self.args.transforms.split(","):
                config[option.strip()] = True

        project = oc4ids.run_transforms(config, self.items(), project_id=project_id)

        if self.args.package:
            output = self.parse_package_arguments()
            output["projects"] = [project]
        else:
            output = project

        self.print(output)



import itertools
from collections import defaultdict
from tempfile import NamedTemporaryFile

from ocdskit.exceptions import InconsistentVersionError, MissingOcidKeyError
from ocdskit.util import (_empty_record_package, _remove_empty_optional_metadata, _resolve_metadata,
                          _update_package_metadata, get_ocds_minor_version, is_release, json_dumps, jsonlib)

try:
    import sqlite3
    using_sqlite = True

    def adapt_json(d):
        return json_dumps(d)

    def convert_json(s):
        return jsonlib.loads(s)

    sqlite3.register_adapter(dict, adapt_json)
    sqlite3.register_converter('json', convert_json)
except ImportError:
    using_sqlite = False


class Packager:
    """
    The Packager context manager helps to build a single record package, or a stream of compiled releases or merged
    releases. Release packages and/or individual releases can be added to the packager. All releases should use the
    same version of OCDS.
    """
    def __enter__(self):
        self.package = _empty_record_package()
        self.version = None

        if using_sqlite:
            self.backend = SQLiteBackend()
        else:
            self.backend = PythonBackend()

        return self

    def __exit__(self, type, value, traceback):
        self.backend.close()

    def add(self, data):
        """
        Adds release packages and/or individual releases to be merged.

        :param data: an iterable of release packages and individual releases
        :raises InconsistentVersionError: if the versions are inconsistent across packages to merge
        """
        for i, item in enumerate(data):
            version = get_ocds_minor_version(item)
            if self.version:
                if version != self.version:
                    # OCDS 1.1 and OCDS 1.0 have different merge rules for `awards.suppliers`. Also, mixing new and
                    # deprecated fields can lead to inconsistencies (e.g. transaction `amount` and `value`).
                    # https://standard.open-contracting.org/latest/en/schema/changelog/#advisories
                    raise InconsistentVersionError('item {}: version error: this item uses version {}, but earlier '
                                                   'items used version {}'.format(i, version, self.version),
                                                   self.version, version)
            else:
                self.version = version

            if is_release(item):
                self.backend.add_release(item, '')
            else:  # release package
                uri = item.get('uri', '')

                _update_package_metadata(self.package, item)

                # Note: If there are millions of packages to merge, we should use SQLite to store the packages instead.
                if uri:
                    self.package['packages'].append(uri)

                for release in item['releases']:
                    self.backend.add_release(release, uri)

            self.backend.flush()

    def output_package(self, merger, return_versioned_release=False, use_linked_releases=False, streaming=False):
        """
        Yields a record package.

        :param ocdsmerge.merge.Merger merger: a merger
        :param bool return_versioned_release: whether to include versioned releases in the record package
        :param bool use_linked_releases: whether to use linked releases instead of full releases, if possible
        :param bool streaming: whether to set the package's records to a generator instead of a list
        """
        records = self.output_records(merger, return_versioned_release=return_versioned_release,
                                      use_linked_releases=use_linked_releases)

        # If a user wants to stream data but can’t exhaust records right away, we can add an `autoclose=True` argument.
        # If set to `False`, `__exit__` will do nothing, and the user will need to call `packager.backend.close()`.
        if not streaming:
            records = list(records)

        self.package['records'] = records

        _resolve_metadata(self.package, 'extensions')
        _remove_empty_optional_metadata(self.package)

        yield self.package

    def output_records(self, merger, return_versioned_release=False, use_linked_releases=False):
        """
        Yields records, ordered by OCID.

        :param ocdsmerge.merge.Merger merger: a merger
        :param bool return_versioned_release: whether to include versioned releases in the record package
        :param bool use_linked_releases: whether to use linked releases instead of full releases, if possible
        """
        for ocid, rows in self.backend.get_releases_by_ocid():
            record = {
                'ocid': ocid,
                'releases': [],
            }

            releases = []
            for _, uri, release in rows:
                releases.append(release)

                if use_linked_releases and uri:
                    package_release = {
                        'url': uri + '#' + release['id'],
                        'date': release['date'],
                        'tag': release['tag'],
                    }
                else:
                    package_release = release
                record['releases'].append(package_release)

            record['compiledRelease'] = merger.create_compiled_release(releases)
            if return_versioned_release:
                record['versionedRelease'] = merger.create_versioned_release(releases)

            yield record

    def output_releases(self, merger, return_versioned_release=False):
        """
        Yields compiled releases or versioned releases, ordered by OCID.

        :param ocdsmerge.merge.Merger merger: a merger
        :param bool return_versioned_release: whether to yield versioned releases instead of compiled releases
        """
        for ocid, rows in self.backend.get_releases_by_ocid():
            releases = (row[-1] for row in rows)

            if return_versioned_release:
                yield merger.create_versioned_release(releases)
            else:
                yield merger.create_compiled_release(releases)


# The backend's responsibilities (for now) are exclusively to:
#
# * Group releases by OCID
# * Store each release's package URI
#
# For a PostgreSQL backend, see https://github.com/open-contracting/ocdskit/issues/116
class AbstractBackend:
    def __init__(self):
        """
        Initializes the backend.
        """
        raise NotImplementedError

    def add_release(self, release, package_uri):
        """
        Adds a release to the backend. (The release might be added to an internal buffer.)

        :raises MissingOcidKeyError: if the release is missing an ``ocid`` key.
        """
        try:
            self._add_release(release['ocid'], package_uri, release)
        except KeyError:
            raise MissingOcidKeyError('ocid')

    def _add_release(self, ocid, package_uri, release):
        raise NotImplementedError

    def get_releases_by_ocid(self):
        """
        Yields an OCIDs and an iterable of tuples of ``(ocid, package_uri, release)``.

        OCIDs are yielded in alphabetical order. The iterable is in any order.
        """
        raise NotImplementedError

    def flush(self):
        """
        Flushes the internal buffer of releases. This may be a no-op on some backends.
        """
        pass

    def close(self):
        """
        Tidies up any resources used by the backend. This may be a no-op on some backends.
        """
        pass


class PythonBackend(AbstractBackend):
    def __init__(self):
        self.groups = defaultdict(list)

    def _add_release(self, ocid, package_uri, release):
        self.groups[ocid].append((ocid, package_uri, release))

    def get_releases_by_ocid(self):
        for ocid in sorted(self.groups):
            yield ocid, self.groups[ocid]


class SQLiteBackend(AbstractBackend):
    # "The sqlite3 module internally uses a statement cache to avoid SQL parsing overhead."
    # https://docs.python.org/3/library/sqlite3.html#sqlite3.connect
    # Note: We never commit changes. SQLite manages the memory usage of uncommitted changes.
    # https://sqlite.org/atomiccommit.html#_cache_spill_prior_to_commit
    def __init__(self):
        self.file = NamedTemporaryFile()

        # https://docs.python.org/3/library/sqlite3.html#sqlite3.PARSE_DECLTYPES
        self.connection = sqlite3.connect(self.file.name, detect_types=sqlite3.PARSE_DECLTYPES)

        # https://sqlite.org/tempfiles.html#temp_databases
        self.connection.execute("CREATE TEMP TABLE releases (ocid text, uri text, release json)")

        self.buffer = []

    def _add_release(self, ocid, package_uri, release):
        self.buffer.append((ocid, package_uri, release))

    def flush(self):
        # https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection.executemany
        self.connection.executemany("INSERT INTO releases VALUES (?, ?, ?)", self.buffer)

        self.buffer = []

    def get_releases_by_ocid(self):
        self.connection.execute("CREATE INDEX IF NOT EXISTS ocid_idx ON releases(ocid)")

        results = self.connection.execute("SELECT * FROM releases ORDER BY ocid")
        for ocid, rows in itertools.groupby(results, lambda row: row[0]):
            yield ocid, rows

    def close(self):
        self.file.close()
        self.connection.close()



import json
from tempfile import TemporaryFile

import pytest

from ocdskit.util import is_record, json_dump
from tests import read


# Same fixture files as in test_detect_format.py, except for concatenated JSON files.
@pytest.mark.parametrize('filename,expected', [
    ('record-package_minimal.json', False),
    ('release-package_minimal.json', False),
    ('record_minimal.json', True),
    ('release_minimal.json', False),
    ('realdata/compiled-release-1.json', False),
    ('realdata/versioned-release-1.json', False),
    ('release-packages.json', False),
    ('detect-format_whitespace.json', False),
])
def test_is_record(filename, expected):
    assert is_record(json.loads(read(filename))) == expected


@pytest.mark.parametrize('data,expected', [
    (iter([]), '[]'),
    ({'a': 1, 'b': 2}, '{"a":1,"b":2}'),
])
def test_json_dump(data, expected):
    with TemporaryFile('w+') as f:
        json_dump(data, f)

        f.seek(0)

        assert f.read() == expected



import logging
import os.path

import pytest

from ocdskit.cli.__main__ import main
from tests import assert_streaming_error, read, run_streaming


@pytest.mark.vcr()
def test_command(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['validate'],
                           ['realdata/release-package-1.json'])

    assert actual == "item 0: 'version' is a required property (required)\n"


@pytest.mark.vcr()
def test_command_invalid_json(monkeypatch, caplog):
    with caplog.at_level(logging.INFO):
        stdin = read('release-package_minimal.json', 'rb') + b'\n{\n'

        assert_streaming_error(monkeypatch, main, ['validate'], stdin)

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'CRITICAL'
        assert caplog.records[0].message.startswith('JSON error: ')


@pytest.mark.vcr()
def test_command_valid_release_package_url(monkeypatch):
    url = 'http://standard.open-contracting.org/schema/1__0__3/release-package-schema.json'

    actual = run_streaming(monkeypatch, main, ['validate', '--schema', url],
                           ['realdata/release-package-1.json'])

    assert actual == ''


@pytest.mark.vcr()
def test_command_valid_release_package_file(monkeypatch):
    url = 'file://{}'.format(os.path.realpath(os.path.join('tests', 'fixtures', 'release-package-schema.json')))

    actual = run_streaming(monkeypatch, main, ['validate', '--schema', url],
                           ['realdata/release-package-1.json'])

    assert actual == ''


@pytest.mark.vcr()
def test_command_valid_release_package_file_verbose(monkeypatch):
    url = 'file://{}'.format(os.path.realpath(os.path.join('tests', 'fixtures', 'release-package-schema.json')))

    actual = run_streaming(monkeypatch, main, ['validate', '--schema', url, '--verbose'],
                           ['realdata/release-package-1.json'])

    assert actual == 'item 0: no errors\n'


@pytest.mark.vcr()
def test_command_invalid_record_package(monkeypatch):
    url = 'https://standard.open-contracting.org/latest/en/record-package-schema.json'

    actual = run_streaming(monkeypatch, main, ['validate', '--schema', url],
                           ['realdata/record-package-1.json'])

    assert "item 0: None is not of type 'string' (properties/records/items/properties/compiledRelease/properties/tender/properties/submissionMethod/items/type)" in actual  # noqa: E501


@pytest.mark.vcr()
def test_command_no_check_urls(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['validate'],
                           ['release-package_urls.json'])

    assert actual == ''


@pytest.mark.vcr()
def test_command_check_urls(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['validate', '--check-urls'],
                           ['release-package_url-error.json'])

    assert actual == """HTTP 500 on GET http://httpbin.org/status/500
item 0: 'http://httpbin.org/status/500' is not a 'uri' (properties/releases/items/properties/tender/properties/documents/items/properties/url/format)
"""  # noqa: E501


# Can't record delay endpoint with VCR.
def test_command_timeout(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['validate', '--check-urls', '--timeout', '1'],
                           ['release-package_url-timeout.json'])

    assert actual == """Timedout on GET http://httpbin.org/delay/3
item 0: 'http://httpbin.org/delay/3' is not a 'uri' (properties/releases/items/properties/tender/properties/documents/items/properties/url/format)
"""  # noqa: E501



import csv
import itertools
import json
import logging
import os.path
from copy import deepcopy

from ocdskit.cli.commands.base import BaseCommand
from ocdskit.util import json_dump

logger = logging.getLogger('ocdskit')


class Command(BaseCommand):
    name = 'set-closed-codelist-enums'
    help = 'sets the enum in a JSON Schema to match the codes in the CSV files of closed codelists'

    def add_arguments(self):
        self.add_argument('standard', help='path to directory containing standard JSON Schema files')
        self.add_argument('extension', help='paths to directories containing extension JSON Schema files', nargs='*')

    def handle(self):
        def collect_codelists(directory):
            for root, dirs, files in os.walk(directory):
                for name in files:
                    if name.endswith('.csv'):
                        with open(os.path.join(root, name)) as f:
                            reader = csv.DictReader(f)
                            row = next(reader)
                            if 'Code' in row:
                                codes = [row['Code'] for row in itertools.chain([row], reader)]
                            else:
                                codes = []

                        if codes:
                            if name.startswith('+'):
                                # KeyError if codelist doesn't exist.
                                codelists[name[1:]] += codes
                            elif name.startswith('-'):
                                # KeyError if codelist doesn't exist.
                                codelists[name[1:]] = [code for code in codelists[name[1:]] if code not in codes]
                            elif name in codelists:
                                if codelists[name] != codes:
                                    logger.error('conflicting codelists: {}'.format(name))
                            else:
                                codelists[name] = codes

        # This method is similar to `validate_codelist_enum` in `test_json.py`.
        def update_codelist_enum(data):
            if isinstance(data, list):
                return [update_codelist_enum(item) for item in data]

            if isinstance(data, dict) and 'codelist' not in data:
                return {key: update_codelist_enum(value) for key, value in data.items()}

            if isinstance(data, dict) and 'codelist' in data:
                codelists_seen.add(data['codelist'])

                if not data['openCodelist']:
                    codes = codelists[data['codelist']]

                    if isinstance(data['type'], str):
                        types = [data['type']]
                    else:
                        types = data['type']

                    if 'string' in types:
                        if 'null' in types:
                            codes.append(None)
                        if 'enum' not in data or set(data['enum']) != set(codes):
                            data['enum'] = codes
                    elif 'array' in types:
                        if 'enum' not in data['items'] or set(data['items']['enum']) != set(codes):
                            data['items']['enum'] = codes

            return data

        def update_json_schema(directory):
            for root, dirs, files in os.walk(directory):
                for name in files:
                    if name.endswith('.json') and name not in meta_schema_exceptions:
                        path = os.path.join(root, name)

                        with open(path) as f:
                            data = json.load(f)

                        # If the JSON file is a JSON Schema file.
                        if any(field in data for field in ('$schema', 'definitions', 'properties')):
                            expected = deepcopy(data)
                            update_codelist_enum(data)

                            if expected != data:
                                with open(path, 'w') as f:
                                    json_dump(data, f, indent=2)
                                    f.write('\n')

        codelists = {}
        codelists_seen = set()
        meta_schema_exceptions = {
            'meta-schema.json',
            'meta-schema-patch.json',
        }

        if self.args.extension:
            collect_codelists(self.args.standard)
            codelists_seen = set(codelists)

            directories = self.args.extension
        else:
            directories = [self.args.standard]

        for directory in directories:
            collect_codelists(directory)
            update_json_schema(directory)

        codelists_not_seen = []
        for codelist in codelists.keys():
            if codelist not in codelists_seen:
                codelists_not_seen.append(codelist)

        if codelists_not_seen:
            logger.error('unused codelists: {}'.format(' '.join(codelists_not_seen)))



from ocdskit.cli.commands.base import OCDSCommand
from ocdskit.combine import combine_record_packages


class Command(OCDSCommand):
    name = 'combine-record-packages'
    help = 'reads record packages from standard input, collects packages and records, and prints one record package'

    def add_arguments(self):
        self.add_package_arguments('record')

    def handle(self):
        kwargs = self.parse_package_arguments()

        output = combine_record_packages(self.items(), **kwargs)

        self.print(output)



import copy
import json

from ocdskit import oc4ids
from tests import read


def test_initial_tranform_state():
    releases = json.loads(read("release-package_additional-contact-points.json"))["releases"]
    transform_state = oc4ids.InitialTransformState(releases, "1")
    assert len(transform_state.compiled_releases) == 1
    assert len(transform_state.releases_by_ocid["ocds-213czf-1"]) == 2


def test_run_all():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["publicAuthority"]}],
        }
    ]
    output = oc4ids.run_transforms({}, releases, "1")
    assert output["parties"] == releases[0]["parties"]


def test_run_all_release_package():
    releases_package_1 = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["publicAuthority"]}],
        }
    ]
    releases_package_2 = [
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "2", "name": "a", "roles": ["publicAuthority"]}],
        }
    ]

    release_packages = [{"uri": "example.com", "releases": releases_package_1},
                        {"uri": "example.com", "releases": releases_package_2}]

    output = oc4ids.run_transforms({}, release_packages, "1")

    assert len(output["contractingProcesses"]) == 2
    assert len(output["parties"]) == 2
    assert output["contractingProcesses"][0]['releases'] == [
        {'url': 'example.com#1', 'date': '2001-02-03T04:05:06Z', 'tag': 'planning'}
    ]


def test_public_authority_role():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [
                {"id": "1", "name": "a", "roles": ["publicAuthority"]},
                {"id": "2", "name": "b", "roles": ["publicAuthority"]},
            ],
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert output["parties"] == releases[0]["parties"]

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["publicAuthority"]}, {"id": "2", "roles": ["buyer"]}],
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 1


def test_duplicate_public_authority_role():

    # Match on identifier
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "a", "roles": ["publicAuthority"], "identifier": {"id": "a", "scheme": "b"}}],
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "a", "roles": ["publicAuthority"], "identifier": {"id": "a", "scheme": "b"}}],
        },
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 1
    assert output["parties"][0]["id"] == "b-a"

    # No match on identifier
    releases[0]["parties"][0]["identifier"]["id"] = "b"

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])

    assert len(output["parties"]) == 2
    print(output["parties"])
    assert output["parties"][0]["id"] == "b-b"
    assert output["parties"][1]["id"] == "b-a"

    # Match on name
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "a", "name": "org 1", "roles": ["publicAuthority"]}],
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "a", "name": "org 1", "roles": ["publicAuthority"]}],
        },
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 1
    assert output["parties"][0]["id"] == "1"

    # No match on name
    releases[0]["parties"][0]["name"] = "org 2"

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 2
    # Generated autoincrement party ids
    assert output["parties"][0]["id"] == "1"
    assert output["parties"][1]["id"] == "2"

    # Match on name but different address
    releases[0]["parties"][0]["name"] = "org 1"
    releases[0]["parties"][0]["address"] = {"streetAddress": "1 the street"}

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 2
    assert output["parties"][0]["id"] == "1"
    assert output["parties"][1]["id"] == "2"

    # Match on name and address
    releases[1]["parties"][0]["address"] = {"streetAddress": "1 the street"}
    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 1
    assert output["parties"][0]["id"] == "1"

    # all roles if match
    releases[1]["parties"][0]["roles"].append("some role")
    releases[1]["parties"][0]["roles"].append("some other role")

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.public_authority_role])
    assert len(output["parties"]) == 1

    assert len(output["parties"][0]["roles"]) == 3
    assert output["parties"][0]["id"] == "1"


def test_buyer_role():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["buyer"]}],
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.buyer_role])

    assert "publicAuthority" in output["parties"][0]["roles"]
    assert "buyer" in output["parties"][0]["roles"]


def test_sector():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "project": {"sector": {"scheme": "COFOG", "description": "Road transportation", "id": "04.5.1"}}
            },
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.sector])
    assert output["sector"] == ["COFOG-04.5.1"]

    # 2 contracting processes but same sector
    releases.append(
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "project": {"sector": {"scheme": "COFOG", "description": "Road transportation", "id": "04.5.1"}}
            },
        }
    )

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.sector])
    assert output["sector"] == ["COFOG-04.5.1"]

    # 2 contracting processes but differnt sector
    releases[1]["planning"]["project"]["sector"]["id"] = "2"
    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.sector])
    assert set(output["sector"]) == set(["COFOG-04.5.1", "COFOG-2"])


def test_additional_classifications():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"additionalClassifications": [{"scheme": "a", "id": "1"}]}},
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.additional_classifications])
    assert output["additionalClassifications"] == [{"scheme": "a", "id": "1"}]

    # same classification
    releases.append(
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"additionalClassifications": [{"scheme": "a", "id": "1"}]}},
        }
    )

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.additional_classifications])
    assert output["additionalClassifications"] == [{"scheme": "a", "id": "1"}]

    # new classification
    releases.append(
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"additionalClassifications": [{"scheme": "a", "id": "2"}]}},
        }
    )
    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.additional_classifications])
    assert output["additionalClassifications"] == [{"scheme": "a", "id": "1"}, {"scheme": "a", "id": "2"}]


def test_title():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"title": "a title"}},
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.title])
    assert output["title"] == releases[0]["planning"]["project"]["title"]

    # clashing titles give warning and no output
    releases.append(
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"title": "b title"}},
        }
    )

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.title])
    assert "title" not in output


def test_title_from_tender():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"title": "a title"},
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.title, oc4ids.title_from_tender])
    assert output["title"] == releases[0]["tender"]["title"]

    releases.append(
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"title": "b title"},
        }
    )

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.title, oc4ids.title_from_tender])

    assert output["title"] == "<ocds-213czf-1> a title\n<ocds-213czf-2> b title\n"

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"title": "a title"}},
            "tender": {"title": "a non used title"},
        }
    ]

    output = oc4ids._run_transforms(releases, "1", transforms=[oc4ids.title, oc4ids.title_from_tender])

    assert output["title"] == releases[0]["planning"]["project"]["title"]


def test_contracting_process_setup_releases():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"title": "a title"},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"title": "a title"},
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup]
    )

    expected = """
    {
      "id": "1",
      "contractingProcesses": [
        {
          "id": "ocds-213czf-1",
          "summary": {
            "ocid": "ocds-213czf-1"
          },
          "embeddedReleases": [
            {
              "ocid": "ocds-213czf-1",
              "id": "1",
              "tag": "planning",
              "date": "2001-02-03T04:05:06Z",
              "tender": {
                "title": "a title"
              }
            }
          ]
        },
        {
          "id": "ocds-213czf-2",
          "summary": {
            "ocid": "ocds-213czf-2"
          },
          "embeddedReleases": [
            {
              "ocid": "ocds-213czf-2",
              "id": "1",
              "tag": "planning",
              "date": "2001-02-03T04:05:06Z",
              "tender": {
                "title": "a title"
              }
            }
          ]
        }
      ]
    }
    """

    assert output == json.loads(expected)


def test_contracting_process_setup_release_packages():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"title": "a title"},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "2",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"title": "a title"},
        },
    ]

    release_packages = [{"uri": "example.com", "releases": releases}]

    output = oc4ids._run_transforms(
        copy.deepcopy(release_packages), "1", transforms=[oc4ids.contracting_process_setup]
    )

    expected = """
    {
      "id": "1",
      "contractingProcesses": [
        {
          "id": "ocds-213czf-1",
          "summary": {
            "ocid": "ocds-213czf-1"
          },
          "releases": [
            {
              "url": "example.com#1",
              "date": "2001-02-03T04:05:06Z",
              "tag": "planning"
            }
          ]
        },
        {
          "id": "ocds-213czf-2",
          "summary": {
            "ocid": "ocds-213czf-2"
          },
          "releases": [
            {
              "url": "example.com#2",
              "date": "2001-02-03T04:05:06Z",
              "tag": "planning"
            }
          ]
        }
      ]
    }
    """

    assert output == json.loads(expected)


def test_procuring_entity():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"procuringEntity": {"id": "1"}},
            "parties": [{"id": "1", "roles": ["procuringEntity"]}],
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.procuring_entity],
    )

    assert output["parties"] == releases[0]["parties"]
    assert output["contractingProcesses"][0]["summary"]["tender"] == releases[0]["tender"]

    # with identifier no duplicate party id
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["procuringEntity"], "identifier": {"id": "a", "scheme": "a"}}],
        },
    ]
    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.procuring_entity],
    )

    assert output["parties"][0]["id"] == "1"
    assert output["contractingProcesses"][0]["summary"]["tender"]["procuringEntity"]["id"] == "1"

    # with identifier and duplicate party id
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["procuringEntity"], "identifier": {"id": "a", "scheme": "a"}}],
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "roles": ["procuringEntity"], "identifier": {"id": "a", "scheme": "a"}}],
        },
    ]
    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.procuring_entity],
    )

    assert output["parties"][0]["id"] == "a-a"
    assert output["contractingProcesses"][0]["summary"]["tender"]["procuringEntity"]["id"] == "a-a"

    # with genderated id
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "name": "org1", "roles": ["procuringEntity"]}],
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "name": "org2", "roles": ["procuringEntity"]}],
        },
    ]
    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.procuring_entity],
    )

    assert output["parties"][0]["id"] == "1"
    assert output["parties"][1]["id"] == "2"
    assert output["contractingProcesses"][0]["summary"]["tender"]["procuringEntity"]["id"] == "1"
    assert output["contractingProcesses"][1]["summary"]["tender"]["procuringEntity"]["id"] == "2"


def test_administrative_entity():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [{"id": "1", "name": "a", "roles": ["administrativeEntity"]}],
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases),
        "1",
        transforms=[oc4ids.contracting_process_setup, oc4ids.administrative_entity],
    )

    assert output["parties"] == releases[0]["parties"]
    assert output["contractingProcesses"][0]["summary"]["tender"]["administrativeEntity"] == {"id": "1", "name": "a"}


def test_multiple_administrative_entity_in_process():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [
                {"id": "1", "name": "a", "roles": ["administrativeEntity"]},
                {"id": "2", "name": "b", "roles": ["administrativeEntity"]},
            ],
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases),
        "1",
        transforms=[oc4ids.contracting_process_setup, oc4ids.administrative_entity],
    )

    assert output["parties"] == releases[0]["parties"]

    # tender is not created as there are multiple adminastrative entities
    assert "tender" not in output["contractingProcesses"][0]["summary"]


def test_contract_status_pre_award():

    releases = [
        {"ocid": "ocds-213czf-1", "id": "1", "tag": "planning", "date": "2001-02-03T04:05:06Z", "tender": {"id": 1}},
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"status": "pending"}],
            "awards": [{"status": "pending"}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-3",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"status": "active"}],
            "awards": [{"status": "pending"}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-4",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "awards": [{"status": "pending", "date": "3000-01-01"}],
            "tender": {"id": 1, "awardPeriod": {"startDate": "3000-01-01"}},
        },
        {
            "ocid": "ocds-213czf-5",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "awards": [{"date": "3000-01-01"}],
            "tender": {"id": 1, "awardPeriod": {"startDate": "2000-01-01"}},
        },
        {
            "ocid": "ocds-213czf-6",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "awards": [{"date": "2000-01-01"}],
            "tender": {"id": 1, "awardPeriod": {"startDate": "3000-01-01"}},
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_status],
    )

    assert output["contractingProcesses"][0]["summary"]["status"] == "pre-award"
    assert output["contractingProcesses"][1]["summary"]["status"] == "pre-award"
    assert output["contractingProcesses"][2]["summary"]["status"] != "pre-award"
    assert output["contractingProcesses"][3]["summary"]["status"] == "pre-award"

    # Currently no status at all as fits no status
    assert output["contractingProcesses"][4]["summary"].get("status") != "pre-award"
    assert output["contractingProcesses"][5]["summary"].get("status") != "pre-award"


def test_contract_status_active():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"status": "active"}, {"status": "pending"}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"startDate": "2000-01-01"}}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-3",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"startDate": "2000-01-01", "endDate": "2000-01-01"}}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-4",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"startDate": "2000-01-01", "endDate": "2000-01-01"}}],
            "awards": [{"contractPeriod": {"startDate": "2000-01-01", "endDate": "3000-01-01"}}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-5",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"startDate": "2000-01-01", "endDate": "2000-01-01"}}],
            "tender": {"id": 1, "contractPeriod": {"startDate": "2000-01-01", "endDate": "3000-01-01"}},
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_status],
    )

    assert output["contractingProcesses"][0]["summary"]["status"] == "active"
    assert output["contractingProcesses"][1]["summary"]["status"] == "active"
    assert output["contractingProcesses"][2]["summary"]["status"] != "active"

    # Currently no status at all as fits no status

    assert output["contractingProcesses"][3]["summary"]["status"] == "active"
    assert output["contractingProcesses"][4]["summary"]["status"] == "active"


def test_contract_status_closed():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"status": "cancelled"},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "awards": [{"status": "cancelled"}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-3",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"status": "cancelled"}],
            "tender": {"id": 1},
        },
        {
            "ocid": "ocds-213czf-4",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"endDate": "2000-01-01"}}],
            "awards": [{"contractPeriod": {"endDate": "2000-01-01"}}],
            "tender": {"id": 1, "contractPeriod": {"endDate": "2000-01-01"}},
        },
        {
            "ocid": "ocds-213czf-5",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"endDate": "3000-01-01"}}],
            "awards": [{"contractPeriod": {"endDate": "2000-01-01"}}],
            "tender": {"id": 1, "contractPeriod": {"endDate": "2000-01-01"}},
        },
        {
            "ocid": "ocds-213czf-6",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"endDate": "2000-01-01"}}],
            "awards": [{"contractPeriod": {"endDate": "3000-01-01"}}],
            "tender": {"id": 1, "contractPeriod": {"endDate": "2000-01-01"}},
        },
        {
            "ocid": "ocds-213czf-7",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"period": {"endDate": "2000-01-01"}}],
            "awards": [{"contractPeriod": {"endDate": "2000-01-01"}}],
            "tender": {"id": 1, "contractPeriod": {"endDate": "3000-01-01"}},
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_status],
    )

    assert output["contractingProcesses"][0]["summary"]["status"] == "closed"
    assert output["contractingProcesses"][1]["summary"]["status"] == "closed"
    assert output["contractingProcesses"][2]["summary"]["status"] == "closed"
    assert output["contractingProcesses"][3]["summary"]["status"] == "closed"
    assert output["contractingProcesses"][4]["summary"]["status"] != "closed"
    assert output["contractingProcesses"][5]["summary"]["status"] != "closed"
    assert output["contractingProcesses"][6]["summary"]["status"] != "closed"


def test_procurment_process():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"procurementMethod": "method", "procurementMethodDetails": "details"},
        }
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases),
        "1",
        transforms=[oc4ids.contracting_process_setup, oc4ids.procurement_process],
    )

    assert output["contractingProcesses"][0]["summary"]["tender"] == releases[0]["tender"]


def test_number_of_tenderers():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"numberOfTenderers": 123},
        }
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases),
        "1",
        transforms=[oc4ids.contracting_process_setup, oc4ids.number_of_tenderers],
    )

    assert output["contractingProcesses"][0]["summary"]["tender"]["numberOfTenderers"] == 123


def test_location():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"locations": [{"description": "Mars"}]}},
        }
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.location],)

    assert output["locations"] == [{"description": "Mars"}]


def test_location_multiple_releases():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"locations": [{"description": "Mars"}, {"description": "Jupiter"}]}},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "2",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"locations": [{"description": "Earth"}]}},
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.location],)

    assert output["locations"] == [{"description": "Mars"}, {"description": "Jupiter"}, {"description": "Earth"}]


def test_location_from_item_location():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {
                "items": [
                    {
                        "id": "item1",
                        "deliveryLocation": {
                            "geometry": {"type": "Point", "coordinates": [51.751944, -1.257778]},
                            "uri": "http://www.geonames.org/2640729/oxford.html",
                        },
                    }
                ],
            },
        }
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.location, oc4ids.location_from_items],
    )
    assert output["locations"] == [releases[0]["tender"]["items"][0]["deliveryLocation"]]


def test_location_from_delivery_address():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {
                "items": [
                    {
                        "id": "item2",
                        "deliveryAddress": {
                            "postalCode": "OX1 1BX",
                            "countryName": "United Kingdom",
                            "streetAddress": "Town Hall, St Aldate's",
                            "region": "Oxfordshire",
                            "locality": "Oxford",
                        },
                    }
                ],
            },
        }
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.location, oc4ids.location_from_items],
    )

    assert output["locations"] == [{"address": releases[0]["tender"]["items"][0]["deliveryAddress"]}]


def test_location_multiple():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {
                "items": [
                    {
                        "id": "item1",
                        "deliveryLocation": {
                            "geometry": {"type": "Point", "coordinates": [51.751944, -1.257778]},
                            "uri": "http://www.geonames.org/2640729/oxford.html",
                        },
                        "deliveryAddress": {
                            "postalCode": "OX1 1BX",
                            "countryName": "United Kingdom",
                            "streetAddress": "Town Hall, St Aldate's",
                            "region": "Oxfordshire",
                            "locality": "Oxford",
                        },
                    }
                ],
            },
        }
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.location, oc4ids.location_from_items],
    )

    assert output["locations"] == [
        releases[0]["tender"]["items"][0]["deliveryLocation"],
        {"address": releases[0]["tender"]["items"][0]["deliveryAddress"]},
    ]


def test_location_not_inferred():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {
                "items": [
                    {
                        "id": "item1",
                        "deliveryLocation": {
                            "geometry": {"type": "Point", "coordinates": [51.751944, -1.257778]},
                            "uri": "http://www.geonames.org/2640729/oxford.html",
                        },
                        "deliveryAddress": {
                            "postalCode": "OX1 1BX",
                            "countryName": "United Kingdom",
                            "streetAddress": "Town Hall, St Aldate's",
                            "region": "Oxfordshire",
                            "locality": "Oxford",
                        },
                    }
                ],
            },
        }
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.location],)

    assert "locations" not in output


def test_budget():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"budget": {"amount": {"amount": "1000", "currency": "USD"}}},
        }
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.budget],)
    assert output["budget"]["amount"] == releases[0]["planning"]["budget"]["amount"]


def test_budget_multiple():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"budget": {"amount": {"amount": "1000", "currency": "USD"}}},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T06:07:08Z",
            "planning": {"budget": {"amount": {"amount": "1234", "currency": "USD"}}},
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.budget],)
    total = float(releases[0]["planning"]["budget"]["amount"]["amount"]) + float(
        releases[1]["planning"]["budget"]["amount"]["amount"]
    )
    assert output["budget"]["amount"]["amount"] == total
    assert output["budget"]["amount"]["currency"] == releases[0]["planning"]["budget"]["amount"]["currency"]


def test_budget_fail():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"budget": {"amount": {"amount": "999", "currency": "USD"}}},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T06:07:08Z",
            "planning": {"budget": {"amount": {"amount": "6464", "currency": "EUR"}}},
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.budget],)
    # Different currencies could not be totalled
    assert "budget" not in output


def test_budget_approval():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "documents": [
                    {"id": "doc1", "documentType": "projectPlan", "title": "A Document"},
                    {"id": "doc2", "documentType": "budgetApproval", "title": "Another Document"},
                ]
            },
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.budget_approval],)
    assert output["documents"] == [releases[0]["planning"]["documents"][1]]

    # duplicate document id in different process, auto increment new doc ids.
    releases.append(
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "documents": [{"id": "doc2", "documentType": "budgetApproval", "title": "Another Another Document"}]
            },
        }
    )

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.budget_approval],)

    assert len(output["documents"]) == 2
    assert output["documents"][0]["id"] == "1"
    assert output["documents"][1]["id"] == "2"


def test_purpose_one():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"rationale": "We were hungry."},
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.purpose],)
    assert output["purpose"] == releases[0]["planning"]["rationale"]


def test_purpose_multiple():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"rationale": "We were hungry."},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "2",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"rationale": "There are never enough post-its."},
        },
    ]

    rationales = "<ocds-213czf-1> We were hungry.\n<ocds-213czf-2> There are never enough post-its.\n"

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.purpose],)
    assert output["purpose"] == rationales


def test_needs_assessment():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "documents": [
                    {"id": "doc1", "documentType": "needsAssessment", "title": "A Document"},
                    {"id": "doc2", "documentType": "budgetApproval", "title": "Another Document"},
                ]
            },
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.purpose_needs_assessment],
    )
    assert output["documents"] == [releases[0]["planning"]["documents"][0]]


def test_description_one():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"description": "A project description"}},
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.description],)
    assert output["description"] == releases[0]["planning"]["project"]["description"]


def test_description_multiple():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"description": "A project description"}},
        },
        {
            "ocid": "ocds-213czf-2",
            "id": "2",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"description": "A project description"}},
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.description],)
    assert output["description"] == releases[0]["planning"]["project"]["description"]

    # contraditing descriptions
    releases[0]["planning"]["project"]["description"] = "another description"

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.description],)

    assert "description" not in output


def test_description_tender():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"description": "A project description"},
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.description, oc4ids.description_tender],
    )
    assert output["description"] == releases[0]["tender"]["description"]

    releases.append(
        {
            "ocid": "ocds-213czf-2",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"description": "A new project description"},
        }
    )
    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.description, oc4ids.description_tender],
    )
    assert (
        output["description"] == "<ocds-213czf-1> A project description\n<ocds-213czf-2> A new project description\n"
    )


def test_description_not_tender():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {"project": {"description": "Another project description"}},
            "tender": {"description": "A project description"},
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.description, oc4ids.description_tender],
    )
    assert output["description"] == releases[0]["planning"]["project"]["description"]


def test_environmental_impact():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "documents": [
                    {"id": "doc1", "documentType": "environmentalImpact", "title": "A Document"},
                    {"id": "doc2", "documentType": "budgetApproval", "title": "Another Document"},
                ]
            },
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.environmental_impact],)
    assert output["documents"] == [releases[0]["planning"]["documents"][0]]


def test_land_and_settlement_impact():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "documents": [
                    {"id": "doc1", "documentType": "environmentalImpact", "title": "A Document"},
                    {"id": "doc2", "documentType": "landAndSettlementImpact", "title": "Another Document"},
                ]
            },
        },
    ]

    output = oc4ids._run_transforms(
        copy.deepcopy(releases), "1", transforms=[oc4ids.land_and_settlement_impact],
    )
    assert output["documents"] == [releases[0]["planning"]["documents"][1]]


def test_project_scope():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "planning": {
                "documents": [
                    {"id": "doc1", "documentType": "projectScope", "title": "A Document"},
                    {"id": "doc2", "documentType": "budgetApproval", "title": "Another Document"},
                ]
            },
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.project_scope],)
    assert output["documents"] == [releases[0]["planning"]["documents"][0]]


def test_project_scope_summary():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {
                "description": "c",
                "items": [{"description": "Some item"}],
                "milestones": [{"description": "Some milestone"}, {"description": "Another milestone"}],
            },
        }
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.project_scope_summary],
    )

    assert "items" in output["contractingProcesses"][0]["summary"]["tender"]
    assert "milestones" in output["contractingProcesses"][0]["summary"]["tender"]
    assert output["contractingProcesses"][0]["summary"]["tender"]["items"] == releases[0]["tender"]["items"]
    assert output["contractingProcesses"][0]["summary"]["tender"]["milestones"] == releases[0]["tender"]["milestones"]


def test_funders_budget():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [
                {
                    "id": "GB-LAC-E09000003-557",
                    "name": "London Borough of Barnet - Transport Services",
                    "details": "This is just a test.",
                },
                {"id": "GB-GOV-23", "name": "Department for Transport", "details": "This is also a test."},
            ],
            "planning": {
                "budget": {
                    "id": "1",
                    "description": "Multi-source budget, see budget breakdown for details.",
                    "amount": {"amount": 300000, "currency": "GBP"},
                    "budgetBreakdown": [
                        {
                            "sourceParty": {
                                "id": "GB-LAC-E09000003-557",
                                "name": "London Borough of Barnet - Transport Services",
                            },
                            "period": {"startDate": "2016-01-01T00:00:00Z", "endDate": "2016-12-31T23:59:59Z"},
                            "id": "001",
                            "description": "Budget contribution from the local government",
                            "amount": {"amount": 150000, "currency": "GBP"},
                        },
                        {
                            "sourceParty": {"id": "GB-GOV-23", "name": "Department for Transport"},
                            "period": {"startDate": "2016-01-01T00:00:00Z", "endDate": "2016-12-31T23:59:59Z"},
                            "id": "002",
                            "description": "Budget contribution from the national government",
                            "amount": {"amount": 150000, "currency": "GBP"},
                        },
                    ],
                }
            },
        }
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.funding_sources],)

    assert output["parties"][0]["id"] == "GB-LAC-E09000003-557"
    assert output["parties"][0]["details"] == "This is just a test."
    assert "funder" in output["parties"][0]["roles"]


def test_funders():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [
                {
                    "id": "GB-LAC-E09000003-557",
                    "name": "London Borough of Barnet - Transport Services",
                    "details": "This is just a test.",
                    "roles": ["funder"],
                },
                {
                    "id": "GB-GOV-23",
                    "name": "Department for Transport",
                    "details": "This is also a test.",
                    "roles": ["funder"],
                },
            ],
        }
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.funding_sources],)

    assert output["parties"][0]["id"] == "GB-LAC-E09000003-557"
    assert output["parties"][0]["details"] == "This is just a test."
    assert "funder" in output["parties"][0]["roles"]


def test_cost_estimate():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "tender": {"status": "planning", "value": {"amount": 1}},
        },
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2100-02-03T04:05:06Z",
            "tender": {"status": "planning", "value": {"amount": 10}},
        },
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.cost_estimate],
    )

    assert output["contractingProcesses"][0]["summary"]["tender"]["costEstimate"] == {"amount": 10}

    # reverse releases
    releases[0]["date"], releases[1]["date"] = releases[1]["date"], releases[0]["date"]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.cost_estimate],
    )

    assert output["contractingProcesses"][0]["summary"]["tender"]["costEstimate"] == {"amount": 1}

    releases.append(
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2200-02-03T04:05:06Z",
            "tender": {"status": "active", "value": {"amount": 100}},
        }
    )

    # last releases is not planning
    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.cost_estimate],
    )
    assert output["contractingProcesses"][0]["summary"]["tender"]["costEstimate"] == {"amount": 1}


def test_contract_title():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"title": "a"}],
            "awards": [{"title": "b"}],
            "tender": {"title": "c"},
        }
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_title],
    )

    assert output["contractingProcesses"][0]["summary"]["title"] == "a"

    # with second contract we use tender title
    releases[0]["contracts"].append({"title": "a"})

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_title],
    )

    assert output["contractingProcesses"][0]["summary"]["title"] == "c"

    # if we remove contracts we use award title
    releases[0].pop("contracts")

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_title],
    )

    assert output["contractingProcesses"][0]["summary"]["title"] == "b"


def test_supplier():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "parties": [
                {"id": "a", "name": "A", "roles": ["supplier"]},
                {"id": "b", "name": "B", "roles": ["supplier"]},
            ],
        }
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.suppliers],
    )

    assert output["parties"] == releases[0]["parties"]

    assert output["contractingProcesses"][0]["summary"]["suppliers"] == [
        {"id": "a", "name": "A"},
        {"id": "b", "name": "B"},
    ]


def test_contract_value():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "awards": [
                {"value": {"amount": 10, "currency": "USD"}},
                {"value": {"amount": 10, "currency": "USD"}},
                {"value": {"amount": 10, "currency": "USD"}},
            ],
        }
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_price],
    )

    assert output["contractingProcesses"][0]["summary"]["contractValue"] == {"amount": 30, "currency": "USD"}

    # change an currency
    releases[0]["awards"][1]["value"]["currency"] = "CAD"

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_price],
    )

    assert "contractValue" not in output["contractingProcesses"][0]["summary"]


def test_contracting_process_description():

    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [{"description": "a", "items": [{"description": "item_a"}]}],
            "awards": [{"description": "b", "items": [{"description": "item_b"}]}],
            "tender": {"description": "c", "items": [{"description": "item_c"}]},
        }
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert output["contractingProcesses"][0]["summary"]["description"] == "a"

    # with no contract description we do not use contract description but item description
    releases[0]["contracts"][0].pop("description")

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert output["contractingProcesses"][0]["summary"]["description"] == "item_a"

    # with no contracts we use awards
    releases[0].pop("contracts")

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert output["contractingProcesses"][0]["summary"]["description"] == "b"

    # with no award description use award item
    releases[0]["awards"][0].pop("description")

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert output["contractingProcesses"][0]["summary"]["description"] == "item_b"

    # with second award item nothing is populated

    releases[0]["awards"][0]["items"].append({"description": "item_b"})

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert "description" not in output["contractingProcesses"][0]["summary"]

    # with a second award uses tender
    releases[0]["awards"].append({"description": "b"})

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert output["contractingProcesses"][0]["summary"]["description"] == "c"

    # with no tender description use items
    releases[0]["tender"].pop("description")

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert output["contractingProcesses"][0]["summary"]["description"] == "item_c"

    # with second tender item we do not have a viable description
    releases[0]["tender"]["items"].append({"description": "item_c"})

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_process_description],
    )

    assert "description" not in output["contractingProcesses"][0]["summary"]


def test_contracting_period():

    releases = [
        {
            "ocid": "ocds-213czf-4",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "awards": [
                {"contractPeriod": {"startDate": "2000-01-01", "endDate": "3000-02-01"}},
                {"contractPeriod": {"startDate": "1999-01-01", "endDate": "3000-01-01"}},
            ],
            "tender": {"contractPeriod": {"startDate": "2100-01-01", "endDate": "2200-01-01"}},
        }
    ]

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_period],
    )

    assert output["contractingProcesses"][0]["summary"]["contractPeriod"] == {
        "startDate": "1999-01-01",
        "endDate": "3000-02-01",
    }

    # remove awards so we get tender contract period
    releases[0].pop("awards")

    output = oc4ids._run_transforms(
        releases, "1", transforms=[oc4ids.contracting_process_setup, oc4ids.contract_period],
    )

    assert output["contractingProcesses"][0]["summary"]["contractPeriod"] == releases[0]["tender"]["contractPeriod"]


def test_final_audit():
    releases = [
        {
            "ocid": "ocds-213czf-1",
            "id": "1",
            "tag": "planning",
            "date": "2001-02-03T04:05:06Z",
            "contracts": [
                {
                    "implementation": {
                        "documents": [
                            {"id": "doc1", "documentType": "finalAudit", "title": "A Document"},
                            {"id": "doc2", "documentType": "budgetApproval", "title": "Another Document"},
                        ]
                    },
                },
                {
                    "implementation": {
                        "documents": [
                            {"id": "doc3", "documentType": "finalAudit", "title": "B Document"},
                            {"id": "doc4", "documentType": "projectScope", "title": "Yet another Document"},
                        ]
                    },
                },
            ],
        },
    ]

    output = oc4ids._run_transforms(copy.deepcopy(releases), "1", transforms=[oc4ids.final_audit],)
    assert output["documents"] == [
        releases[0]["contracts"][0]["implementation"]["documents"][0],
        releases[0]["contracts"][1]["implementation"]["documents"][0],
    ]



import json

from ocdskit.cli.__main__ import main
from tests import assert_streaming, run_streaming


def test_command(monkeypatch):
    assert_streaming(monkeypatch, main, ['combine-record-packages'],
                     ['record-package_minimal.json', 'record-package_maximal.json', 'record-package_extensions.json'],
                     ['combine-record-packages_minimal-maximal-extensions.json'])


def test_command_no_extensions(monkeypatch):
    assert_streaming(monkeypatch, main, ['combine-record-packages'],
                     ['record-package_minimal.json'],
                     ['combine-record-packages_minimal.json'])


def test_command_uri_published_date(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['combine-record-packages', '--uri', 'http://example.com/x.json',
                                               '--published-date', '2010-01-01T00:00:00Z'],
                           ['record-package_minimal.json'])

    package = json.loads(actual)
    assert package['uri'] == 'http://example.com/x.json'
    assert package['publishedDate'] == '2010-01-01T00:00:00Z'


def test_command_publisher(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['combine-record-packages', '--publisher-name', 'Acme Inc.'],
                           ['record-package_minimal.json'])

    package = json.loads(actual)
    assert package['publisher']['name'] == 'Acme Inc.'


def test_command_packages(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['combine-record-packages'],
                           ['realdata/record-package_package.json', 'realdata/record-package_package.json'])

    package = json.loads(actual)
    assert package['packages'] == [
        'http://www.contratosabiertos.cdmx.gob.mx/api/contrato/OCDS-87SD3T-AD-SF-DRM-063-2015',
        'http://www.contratosabiertos.cdmx.gob.mx/api/contrato/OCDS-87SD3T-AD-SF-DRM-065-2015',
    ]



import json

from ocdskit.cli.__main__ import main
from tests import run_streaming

stdin = b'''{
  "required": [
    "array",
    "minItemsArray",
    "object",
    "minPropertiesObject",
    "string",
    "minLengthString",
    "enumString",
    "formatString",
    "patternString"
  ],
  "properties": {
    "optionalArray": {
      "type": "array"
    },
    "array": {
      "type": "array"
    },
    "minItemsArray": {
      "type": "array",
      "minItems": 2
    },
    "optionalObject": {
      "type": "object"
    },
    "object": {
      "type": "object"
    },
    "minPropertiesObject": {
      "type": "object",
      "minProperties": 2
    },
    "optionalString": {
      "type": "string"
    },
    "string": {
      "type": "string"
    },
    "minLengthString": {
      "type": "string",
      "minLength": 2
    },
    "enumString": {
      "type": "string",
      "enum": [
        "a"
      ]
    },
    "formatString": {
      "type": "string",
      "format": "uri"
    },
    "patternString": {
      "type": "string",
      "pattern": "."
    }
  }
}
'''

expected = '''{
  "required": [
    "array",
    "minItemsArray",
    "object",
    "minPropertiesObject",
    "string",
    "minLengthString",
    "enumString",
    "formatString",
    "patternString"
  ],
  "properties": {
    "optionalArray": {
      "type": "array",
      "uniqueItems": true
    },
    "array": {
      "type": "array",
      "minItems": 1,
      "uniqueItems": true
    },
    "minItemsArray": {
      "type": "array",
      "minItems": 2,
      "uniqueItems": true
    },
    "optionalObject": {
      "type": "object"
    },
    "object": {
      "type": "object",
      "minProperties": 1
    },
    "minPropertiesObject": {
      "type": "object",
      "minProperties": 2
    },
    "optionalString": {
      "type": "string"
    },
    "string": {
      "type": "string",
      "minLength": 1
    },
    "minLengthString": {
      "type": "string",
      "minLength": 2
    },
    "enumString": {
      "type": "string",
      "enum": [
        "a"
      ]
    },
    "formatString": {
      "type": "string",
      "format": "uri"
    },
    "patternString": {
      "type": "string",
      "pattern": "."
    }
  }
}
'''


def test_command(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['--pretty', 'schema-strict'], stdin)

    assert actual == expected


def test_command_no_unique_items(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['schema-strict', '--no-unique-items'], stdin)

    assert 'uniqueItems' not in json.loads(actual)['properties']['array']



import os.path
from tempfile import TemporaryDirectory

import pytest

from ocdskit.cli.__main__ import main
from tests import assert_command, path, run_command

test_command_argvalues = [
    ('record-package_minimal.json', 'record package'),
    ('release-package_minimal.json', 'release package'),
    ('record_minimal.json', 'record'),
    ('release_minimal.json', 'release'),
    ('realdata/compiled-release-1.json', 'compiled release'),
    ('realdata/versioned-release-1.json', 'versioned release'),
    ('release-packages.json', 'a JSON array of release packages'),
    ('release-packages.jsonl', 'concatenated JSON, starting with a JSON array of release packages'),
    ('detect-format_mixed.json', 'concatenated JSON, starting with release'),
    ('detect-format_whitespace.json', 'release'),
]

test_command_unknown_format_argvalues = [
    ('false', 'boolean'),
    ('null', 'null'),
    ('number', 'number'),
    ('string', 'string'),
    ('true', 'boolean'),
    ('array', 'non-OCDS array'),
    ('object', 'non-OCDS object'),
]

content = b'{"records":[]}'


@pytest.mark.parametrize('filename,result', test_command_argvalues)
def test_command(filename, result, monkeypatch):
    expected = 'tests/fixtures/{}: {}\n'.format(filename, result)
    assert_command(monkeypatch, main, ['detect-format', path(filename)], expected)


def test_command_recursive(monkeypatch, caplog):
    with TemporaryDirectory() as d:
        with open(os.path.join(d, 'test.json'), 'wb') as f:
            f.write(content)

        with open(os.path.join(d, '.test.json'), 'wb') as f:
            f.write(content)

        actual = run_command(monkeypatch, main, ['detect-format', '--recursive', d])

        assert '/test.json: record package' in actual
        assert '.test.json' not in actual
        assert len(caplog.records) == 0


@pytest.mark.parametrize('basename,result', test_command_unknown_format_argvalues)
def test_command_unknown_format(basename, result, monkeypatch, caplog):
    filename = 'detect-format_{}.json'.format(basename)

    assert_command(monkeypatch, main, ['detect-format', path(filename)], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message == 'tests/fixtures/{}: unknown (top-level JSON value is a {})'.format(
        filename, result)


def test_command_directory(monkeypatch, caplog):
    with TemporaryDirectory() as d:
        assert_command(monkeypatch, main, ['detect-format', d], '')

        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == 'WARNING'
        assert caplog.records[0].message.endswith(' is a directory. Set --recursive to recurse into directories.')


def test_command_nonexistent(monkeypatch, caplog):
    assert_command(monkeypatch, main, ['detect-format', 'nonexistent'], '')

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == 'ERROR'
    assert caplog.records[0].message == 'nonexistent: No such file or directory'



from ocdskit.cli.__main__ import main
from tests import assert_streaming


def test_command(monkeypatch):
    assert_streaming(monkeypatch, main, ['package-releases', '--uri', 'http://example.com', '--published-date',
                                         '9999-01-01T00:00:00Z', '--publisher-name', ''],
                     ['release_minimal-1.json', 'release_minimal-2.json'],
                     ['release-package_minimal-1-2.json'])


def test_command_size(monkeypatch):
    assert_streaming(monkeypatch, main, ['package-releases', '--size', '2'],
                     ['release_minimal-1.json', 'release_minimal-2.json', 'release_minimal.json'],
                     ['release-package_minimal-1-2-no-metadata.json', 'release-package_minimal-no-metadata.json'])


def test_command_extensions(monkeypatch):
    assert_streaming(monkeypatch, main, ['package-releases', '--uri', 'http://example.com', '--published-date',
                                         '9999-01-01T00:00:00Z', '--publisher-name', '',
                                         'http://example.com/a/extension.json', 'http://example.com/b/extension.json'],
                     ['release_minimal-1.json', 'release_minimal-2.json'],
                     ['release-package_minimal-1-2-extensions.json'])


def test_command_root_path_array(monkeypatch):
    assert_streaming(monkeypatch, main, ['package-releases', '--root-path', 'records.item.releases'],
                     ['realdata/record-package-1.json', 'realdata/record-package-2.json'],
                     ['realdata/release-package_record-package.json'])


def test_command_root_path_item(monkeypatch):
    assert_streaming(monkeypatch, main, ['package-releases', '--root-path', 'records.item.releases.item'],
                     ['realdata/record-package-1.json', 'realdata/record-package-2.json'],
                     ['realdata/release-package_record-package.json'])



import pytest

from ocdskit.cli.__main__ import main
from tests import run_streaming


@pytest.mark.vcr()
def test_command(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['tabulate', 'sqlite://'], ['realdata/release-package-1.json'])

    assert actual == ''


@pytest.mark.vcr()
def test_command_release_package(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['tabulate', 'sqlite://'], ['release-package_minimal.json'])

    assert actual == ''


@pytest.mark.vcr()
def test_command_record_package(monkeypatch):
    actual = run_streaming(monkeypatch, main, ['tabulate', 'sqlite://'], ['record-package_minimal.json'])

    assert actual == ''


@pytest.mark.vcr()
def test_command_drop(monkeypatch, tmpdir):
    stdin = ['release-package_minimal.json']
    database_url = 'sqlite:///{}'.format(tmpdir.join('tmp.db'))

    run_streaming(monkeypatch, main, ['tabulate', database_url], stdin)
    actual = run_streaming(monkeypatch, main, ['tabulate', database_url, '--drop'], stdin)

    assert actual == ''
